import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from tonic.tensorflow import updaters

FLOAT_EPSILON = 1e-8


class StochasticPolicyGradient:
    def __init__(self, optimizer=None, entropy_coeff=0, gradient_clip=0):
        self.optimizer = optimizer or tf.keras.optimizers.Adam(
            lr=3e-4, epsilon=1e-8
        )
        self.entropy_coeff = entropy_coeff
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = self.model.actor.trainable_variables

    @tf.function
    def __call__(self, observations, actions, advantages, log_probs):
        if tf.math.count_nonzero(advantages) == 0:
            loss, kl = 0.0, 0.0
            distributions = self.model.actor(observations)
            entropy = tf.reduce_mean(distributions.entropy())
            std = tf.reduce_mean(distributions.stddev())

        else:
            with tf.GradientTape() as tape:
                distributions = self.model.actor(observations)
                new_log_probs = distributions.log_prob(actions)
                loss = -tf.reduce_mean(advantages * new_log_probs)
                entropy = tf.reduce_mean(distributions.entropy())
                if self.entropy_coeff != 0:
                    loss -= self.entropy_coeff * entropy

            gradients = tape.gradient(loss, self.variables)
            if self.gradient_clip > 0:
                gradients = tf.clip_by_global_norm(
                    gradients, self.gradient_clip
                )[0]
            self.optimizer.apply_gradients(zip(gradients, self.variables))

            kl = tf.reduce_mean(log_probs - new_log_probs)
            std = tf.reduce_mean(distributions.stddev())

        return dict(loss=loss, kl=kl, entropy=entropy, std=std)


class ClippedRatio:
    def __init__(
        self,
        optimizer=None,
        ratio_clip=0.2,
        kl_threshold=0.015,
        entropy_coeff=0,
        gradient_clip=0,
    ):
        self.optimizer = optimizer or tf.keras.optimizers.Adam(
            lr=3e-4, epsilon=1e-8
        )
        self.ratio_clip = ratio_clip
        self.kl_threshold = kl_threshold
        self.entropy_coeff = entropy_coeff
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = self.model.actor.trainable_variables

    @tf.function
    def __call__(self, observations, actions, advantages, log_probs):
        if tf.math.count_nonzero(advantages) == 0:
            loss, kl, clip_fraction = 0.0, 0.0, 0.0
            distributions = self.model.actor(observations)
            entropy = tf.reduce_mean(distributions.entropy())
            std = tf.reduce_mean(distributions.stddev())

        else:
            with tf.GradientTape() as tape:
                distributions = self.model.actor(observations)
                new_log_probs = distributions.log_prob(actions)
                ratios_1 = tf.exp(new_log_probs - log_probs)
                surrogates_1 = advantages * ratios_1
                ratio_low = 1 - self.ratio_clip
                ratio_high = 1 + self.ratio_clip
                ratios_2 = tf.clip_by_value(ratios_1, ratio_low, ratio_high)
                surrogates_2 = advantages * ratios_2
                loss = -tf.reduce_mean(tf.minimum(surrogates_1, surrogates_2))
                entropy = tf.reduce_mean(distributions.entropy())
                if self.entropy_coeff != 0:
                    loss -= self.entropy_coeff * entropy

            gradients = tape.gradient(loss, self.variables)
            if self.gradient_clip > 0:
                gradients = tf.clip_by_global_norm(
                    gradients, self.gradient_clip
                )[0]
            self.optimizer.apply_gradients(zip(gradients, self.variables))

            kl = tf.reduce_mean(log_probs - new_log_probs)
            clipped = tf.logical_or(
                ratios_1 > ratio_high, ratios_1 < ratio_low
            )
            clip_fraction = tf.reduce_mean(tf.cast(clipped, tf.float32))
            std = tf.reduce_mean(distributions.stddev())

        return dict(
            loss=loss,
            kl=kl,
            entropy=entropy,
            clip_fraction=clip_fraction,
            std=std,
            stop=kl > self.kl_threshold,
        )


class TrustRegionPolicyGradient:
    def __init__(self, optimizer=None, entropy_coeff=0):
        self.optimizer = optimizer or updaters.ConjugateGradient()
        self.entropy_coeff = entropy_coeff

    def initialize(self, model):
        self.model = model
        self.variables = self.model.actor.trainable_variables

    # Careful: do not use in @tf.function
    def __call__(
        self, observations, actions, log_probs, locs, scales, advantages
    ):
        if np.all(advantages == 0.0):
            kl = tf.convert_to_tensor(0.0, dtype=tf.float32)
            loss = tf.convert_to_tensor(0.0, dtype=tf.float32)
            steps = tf.convert_to_tensor(0, dtype=tf.float32)

        else:
            kl, loss, steps = self.optimizer.optimize(
                loss_function=lambda: self._loss(
                    observations, actions, log_probs, advantages
                ),
                constraint_function=lambda: self._kl(
                    observations, locs, scales
                ),
                variables=self.variables,
            )

        return dict(loss=loss, kl=kl, backtrack_steps=steps)

    @tf.function
    def _loss(self, observations, actions, old_log_probs, advantages):
        distributions = self.model.actor(observations)
        log_probs = distributions.log_prob(actions)
        ratios = tf.exp(log_probs - old_log_probs)
        loss = -tf.reduce_mean(ratios * advantages)
        if self.entropy_coeff != 0:
            entropy = tf.reduce_mean(distributions.entropy())
            loss -= self.entropy_coeff * entropy
        return loss

    @tf.function
    def _kl(self, observations, locs, scales):
        distributions = self.model.actor(observations)
        old_distributions = type(distributions)(locs, scales)
        return tf.reduce_mean(distributions.kl_divergence(old_distributions))


class DeterministicPolicyGradient:
    def __init__(self, optimizer=None, gradient_clip=0):
        self.optimizer = optimizer or tf.keras.optimizers.Adam(
            lr=1e-3, epsilon=1e-8
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = self.model.actor.trainable_variables

    @tf.function
    def __call__(self, observations):
        with tf.GradientTape() as tape:
            actions = self.model.actor(observations)
            values = self.model.critic(observations, actions)
            loss = -tf.reduce_mean(values)

        gradients = tape.gradient(loss, self.variables)
        if self.gradient_clip > 0:
            gradients = tf.clip_by_global_norm(gradients, self.gradient_clip)[
                0
            ]
        self.optimizer.apply_gradients(zip(gradients, self.variables))

        return dict(loss=loss)


class DistributionalDeterministicPolicyGradient:
    def __init__(self, optimizer=None, gradient_clip=0):
        self.optimizer = optimizer or tf.keras.optimizers.Adam(
            lr=1e-3, epsilon=1e-8
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = self.model.actor.trainable_variables

    @tf.function
    def __call__(self, observations):
        with tf.GradientTape() as tape:
            actions = self.model.actor(observations)
            value_distributions = self.model.critic(observations, actions)
            values = value_distributions.mean()
            loss = -tf.reduce_mean(values)

        gradients = tape.gradient(loss, self.variables)
        if self.gradient_clip > 0:
            gradients = tf.clip_by_global_norm(gradients, self.gradient_clip)[
                0
            ]
        self.optimizer.apply_gradients(zip(gradients, self.variables))

        return dict(loss=loss)


class TwinCriticSoftDeterministicPolicyGradient:
    def __init__(self, optimizer=None, entropy_coeff=0.2, gradient_clip=0):
        self.optimizer = optimizer or tf.keras.optimizers.Adam(
            lr=1e-3, epsilon=1e-8
        )
        self.entropy_coeff = entropy_coeff
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = self.model.actor.trainable_variables

    @tf.function
    def __call__(self, observations):
        with tf.GradientTape() as tape:
            distributions = self.model.actor(observations)
            if hasattr(distributions, "sample_with_log_prob"):
                actions, log_probs = distributions.sample_with_log_prob()
            else:
                actions = distributions.sample()
                log_probs = distributions.log_prob(actions)
            values_1 = self.model.critic_1(observations, actions)
            values_2 = self.model.critic_2(observations, actions)
            values = tf.minimum(values_1, values_2)
            loss = tf.reduce_mean(self.entropy_coeff * log_probs - values)

        gradients = tape.gradient(loss, self.variables)
        if self.gradient_clip > 0:
            gradients = tf.clip_by_global_norm(gradients, self.gradient_clip)[
                0
            ]
        self.optimizer.apply_gradients(zip(gradients, self.variables))

        return dict(loss=loss)


class MaximumAPosterioriPolicyOptimization:
    def __init__(
        self,
        num_samples=20,
        epsilon=1e-1,
        epsilon_penalty=1e-3,
        epsilon_mean=1e-3,
        epsilon_std=1e-6,
        initial_log_temperature=1.0,
        initial_log_alpha_mean=1.0,
        initial_log_alpha_std=10.0,
        min_log_dual=-18.0,
        per_dim_constraining=True,
        action_penalization=True,
        actor_optimizer=None,
        dual_optimizer=None,
        gradient_clip=0,
    ):
        self.num_samples = num_samples
        self.epsilon = epsilon
        self.epsilon_mean = epsilon_mean
        self.epsilon_std = epsilon_std
        self.initial_log_temperature = initial_log_temperature
        self.initial_log_alpha_mean = initial_log_alpha_mean
        self.initial_log_alpha_std = initial_log_alpha_std
        self.min_log_dual = min_log_dual
        self.action_penalization = action_penalization
        self.epsilon_penalty = epsilon_penalty
        self.per_dim_constraining = per_dim_constraining
        self.actor_optimizer = actor_optimizer or tf.keras.optimizers.Adam(
            lr=3e-4, epsilon=1e-8
        )
        self.dual_optimizer = actor_optimizer or tf.keras.optimizers.Adam(
            lr=1e-2, epsilon=1e-8
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model, action_space):
        self.model = model
        self.actor_variables = model.actor.trainable_variables

        # Dual variables.
        self.dual_variables = []
        self.log_temperature = tf.Variable(
            [self.initial_log_temperature], dtype=tf.float32
        )
        self.dual_variables.append(self.log_temperature)
        shape = [action_space.shape[0]] if self.per_dim_constraining else [1]
        self.log_alpha_mean = tf.Variable(
            tf.fill(shape, self.initial_log_alpha_mean), dtype=tf.float32
        )
        self.dual_variables.append(self.log_alpha_mean)
        self.log_alpha_std = tf.Variable(
            tf.fill(shape, self.initial_log_alpha_std), dtype=tf.float32
        )
        self.dual_variables.append(self.log_alpha_std)
        if self.action_penalization:
            self.log_penalty_temperature = tf.Variable(
                [self.initial_log_temperature], dtype=tf.float32
            )
            self.dual_variables.append(self.log_penalty_temperature)

    @tf.function
    def __call__(self, observations):
        def parametric_kl_and_dual_losses(kl, alpha, epsilon):
            kl_mean = tf.reduce_mean(kl, axis=0)
            kl_loss = tf.reduce_sum(tf.stop_gradient(alpha) * kl_mean)
            alpha_loss = tf.reduce_sum(
                alpha * (epsilon - tf.stop_gradient(kl_mean))
            )
            return kl_loss, alpha_loss

        def weights_and_temperature_loss(q_values, epsilon, temperature):
            tempered_q_values = tf.stop_gradient(q_values) / temperature
            weights = tf.nn.softmax(tempered_q_values, axis=0)
            weights = tf.stop_gradient(weights)

            # Temperature loss (dual of the E-step).
            q_log_sum_exp = tf.reduce_logsumexp(tempered_q_values, axis=0)
            num_actions = tf.cast(q_values.shape[0], tf.float32)
            log_num_actions = tf.math.log(num_actions)
            loss = epsilon + tf.reduce_mean(q_log_sum_exp) - log_num_actions
            loss = temperature * loss

            return weights, loss

        # Use independent normals to satisfy KL constraints per-dimension.
        def independent_normals(distribution_1, distribution_2=None):
            distribution_2 = distribution_2 or distribution_1
            return tfp.distributions.Independent(
                tfp.distributions.Normal(
                    distribution_1.mean(), distribution_2.stddev()
                )
            )

        self.log_temperature.assign(
            tf.maximum(self.min_log_dual, self.log_temperature)
        )
        self.log_alpha_mean.assign(
            tf.maximum(self.min_log_dual, self.log_alpha_mean)
        )
        self.log_alpha_std.assign(
            tf.maximum(self.min_log_dual, self.log_alpha_std)
        )
        if self.action_penalization:
            self.log_penalty_temperature.assign(
                tf.maximum(self.min_log_dual, self.log_penalty_temperature)
            )

        target_distributions = self.model.target_actor(observations)
        actions = target_distributions.sample(self.num_samples)

        tiled_observations = updaters.tile(observations, self.num_samples)
        flat_observations = updaters.merge_first_two_dims(tiled_observations)
        flat_actions = updaters.merge_first_two_dims(actions)
        values = self.model.target_critic(flat_observations, flat_actions)
        values = tf.reshape(values, (self.num_samples, -1))

        assert isinstance(
            target_distributions, tfp.distributions.MultivariateNormalDiag
        )
        target_distributions = independent_normals(target_distributions)

        with tf.GradientTape() as tape:
            distributions = self.model.actor(observations)
            distributions = independent_normals(distributions)

            temperature = (
                tf.math.softplus(self.log_temperature) + FLOAT_EPSILON
            )
            alpha_mean = tf.math.softplus(self.log_alpha_mean) + FLOAT_EPSILON
            alpha_std = tf.math.softplus(self.log_alpha_std) + FLOAT_EPSILON
            weights, temperature_loss = weights_and_temperature_loss(
                values, self.epsilon, temperature
            )

            # Action penalization is quadratic beyond [-1, 1].
            if self.action_penalization:
                penalty_temperature = (
                    tf.math.softplus(self.log_penalty_temperature)
                    + FLOAT_EPSILON
                )
                diff_bounds = actions - tf.clip_by_value(actions, -1, 1)
                action_bound_costs = -tf.norm(diff_bounds, axis=-1)
                (
                    penalty_weights,
                    penalty_temperature_loss,
                ) = weights_and_temperature_loss(
                    action_bound_costs,
                    self.epsilon_penalty,
                    penalty_temperature,
                )
                weights += penalty_weights
                temperature_loss += penalty_temperature_loss

            # Decompose the policy into fixed-mean and fixed-std distributions.
            fixed_std_distribution = independent_normals(
                distributions.distribution, target_distributions.distribution
            )
            fixed_mean_distribution = independent_normals(
                target_distributions.distribution, distributions.distribution
            )

            # Compute the decomposed policy losses.
            policy_mean_losses = tf.reduce_sum(
                fixed_std_distribution.log_prob(actions) * weights, axis=0
            )
            policy_mean_loss = -tf.reduce_mean(policy_mean_losses)
            policy_std_losses = tf.reduce_sum(
                fixed_mean_distribution.log_prob(actions) * weights, axis=0
            )
            policy_std_loss = -tf.reduce_mean(policy_std_losses)

            # Compute the decomposed KL between the target and online policies.
            if self.per_dim_constraining:
                kl_mean = target_distributions.distribution.kl_divergence(
                    fixed_std_distribution.distribution
                )
                kl_std = target_distributions.distribution.kl_divergence(
                    fixed_mean_distribution.distribution
                )
            else:
                kl_mean = target_distributions.kl_divergence(
                    fixed_std_distribution
                )
                kl_std = target_distributions.kl_divergence(
                    fixed_mean_distribution
                )

            # Compute the alpha-weighted KL-penalty and dual losses.
            kl_mean_loss, alpha_mean_loss = parametric_kl_and_dual_losses(
                kl_mean, alpha_mean, self.epsilon_mean
            )
            kl_std_loss, alpha_std_loss = parametric_kl_and_dual_losses(
                kl_std, alpha_std, self.epsilon_std
            )

            # Combine losses.
            policy_loss = policy_mean_loss + policy_std_loss
            kl_loss = kl_mean_loss + kl_std_loss
            dual_loss = alpha_mean_loss + alpha_std_loss + temperature_loss
            loss = policy_loss + kl_loss + dual_loss

        actor_gradients, dual_gradients = tape.gradient(
            loss, (self.actor_variables, self.dual_variables)
        )

        if self.gradient_clip > 0:
            actor_gradients = tf.clip_by_global_norm(
                actor_gradients, self.gradient_clip
            )[0]
            dual_gradients = tf.clip_by_global_norm(
                dual_gradients, self.gradient_clip
            )[0]

        self.actor_optimizer.apply_gradients(
            zip(actor_gradients, self.actor_variables)
        )
        self.dual_optimizer.apply_gradients(
            zip(dual_gradients, self.dual_variables)
        )

        dual_variables = dict(
            temperature=temperature, alpha_mean=alpha_mean, alpha_std=alpha_std
        )
        if self.action_penalization:
            dual_variables["penalty_temperature"] = penalty_temperature

        return dict(
            policy_mean_loss=policy_mean_loss,
            policy_std_loss=policy_std_loss,
            kl_mean_loss=kl_mean_loss,
            kl_std_loss=kl_std_loss,
            alpha_mean_loss=alpha_mean_loss,
            alpha_std_loss=alpha_std_loss,
            temperature_loss=temperature_loss,
            **dual_variables,
        )



"""Script used to plot results."""

import argparse
import itertools
import os
import pathlib
import time

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import gridspec, lines

from deprl.vendor.tonic import logger


def smooth(vals, window):
    """Smooths values using a sliding window."""

    if window > 1:
        if window > len(vals):
            window = len(vals)
        y = np.ones(window)
        x = vals
        z = np.ones(len(vals))
        mode = "same"
        vals = np.convolve(x, y, mode) / np.convolve(z, y, mode)

    return vals


def stats(xs, means, stds):
    """Extract statistics over the common range of x values."""

    # Extract the common range of x values.
    lengths = [len(x) for x in xs]
    min_length = min(lengths)
    xs = [x[:min_length] for x in xs]
    assert len(xs) == 1 or np.all(xs[1:] == xs[0]), xs
    x = xs[0]

    # Compute statistics.
    means = np.array([mean[:min_length] for mean in means])
    mean = means.mean(axis=0)
    min_mean = means.min(axis=0)
    max_mean = means.max(axis=0)
    if stds is not None:
        stds = np.array([std[:min_length] for std in stds])
        variances = stds**2
        var_within = variances.mean(axis=0)
        var_between = ((means - mean) ** 2).mean(axis=0)
        std = np.sqrt(var_within + var_between)
    else:
        std = None

    return x, mean, min_mean, max_mean, std


def flip(items, columns):
    """Flips values to fill horizontally the legend instead of vertically."""

    return itertools.chain(*[items[i::columns] for i in range(columns)])


def get_data(
    paths, baselines, baselines_source, x_axis, y_axis, x_min, x_max, window
):
    """Gets data from the paths and benchmark."""

    data = {}

    # List the log files in paths.
    log_paths = []
    for path in paths:
        if os.path.isdir(path):
            log_paths.extend(pathlib.Path(path).rglob("log.*"))
        elif path[-7:-3] == "log.":
            log_paths.append(path)

    # Load the data from the log files.
    for path in log_paths:
        sub_path, file = os.path.split(path)
        dfs = {}

        if file == "log.csv":
            # Extract the environment, agent and seed from the paths.
            env, agent, seed = sub_path.split(os.sep)[-3:]

            # Load the data.
            df_seed = pd.read_csv(path, sep=",", engine="python")
            x = df_seed[x_axis].values
            # The x axis should be sorted.
            if not np.all(np.diff(x) > 0):
                logger.warning(f"Skipping unsorted {env} {agent} {seed}")
                continue
            if x_min and x[-1] < x_min:
                logger.warning(
                    f"Skipping {env} {agent} {seed} ({x[-1]} steps)"
                )
                continue
            dfs[seed] = df_seed

        elif file == "log.pkl":
            # Extract the environment, agent and seed from the paths.
            env, agent = sub_path.split(os.sep)[-2:]

            # Load the data.
            df = pd.read_pickle(path, compression="zip")
            for seed, df_seed in df.groupby("seed"):
                x = df_seed[x_axis].values
                # The x axis should be sorted.
                if not np.all(np.diff(x) > 0):
                    logger.warning(f"Skipping unsorted {env} {agent} {seed}")
                    continue
                if x_min and x[-1] < x_min:
                    logger.warning(
                        f"Skipping {env} {agent} {seed} ({x[-1]} steps)"
                    )
                    continue
                dfs[seed] = df_seed

        for seed, df in dfs.items():
            if env not in data:
                data[env] = {}
            if agent not in data[env]:
                data[env][agent] = {}
            assert seed not in data[env][agent]

            # Extract the data.
            x = df[x_axis].values
            if y_axis in df:
                mean = df[y_axis].values
                if y_axis[-5:] == "/mean" and y_axis[:-5] + "/std" in df:
                    std = df[y_axis[:-5] + "/std"].values
                else:
                    std = None
            elif y_axis + "/mean" in df:
                mean = df[y_axis + "/mean"].values
                if y_axis + "/std" in df:
                    std = df[y_axis + "/std"].values
                else:
                    std = None
            else:
                raise KeyError(f"Key {y_axis} not found.")

            # Limit the range of x values and smooth.
            if x_max:
                max_index = np.argmax(x > x_max) or len(x)
            else:
                max_index = len(x)
            x = x[:max_index]
            mean = smooth(mean, window)
            mean = mean[:max_index]
            if std is not None:
                std = smooth(std, window)
                std = std[:max_index]

            data[env][agent][seed] = x, mean, std

    # Aggregate the runs and compute the statistics.
    for env, env_data in data.items():
        for agent, agent_data in env_data.items():
            xs, means, stds = [], [], []
            for seed, (x, mean, std) in agent_data.items():
                xs.append(x)
                means.append(mean)
                stds.append(std)
            if stds[0] is None:
                stds = None
            print(env, agent)
            env_data[agent] = dict(
                seeds=(xs, means), stats=stats(xs, means, stds)
            )

    # Load the data from the benchmark if needed.
    if baselines:
        full_benchmark = len(paths) == 0
        path = __file__
        tonic_path = os.path.split(os.path.split(path)[0])[0]

        # Find the benchmark data to use.
        if baselines_source == "tensorflow":
            logger.log("Loading TensorFlow baselines...")
            path = os.path.join(tonic_path, "data/logs/tensorflow_logs.pkl")
            print("Path:", path)
        else:
            logger.log(f"Loading baselines from {baselines_source}...")
            path = baselines_source

        # Load the compressed benchmark data.
        df = pd.read_pickle(path, compression="zip")

        # Search for data on the corresponding environments and agents.
        for env, df_env in df.groupby("environment"):
            if full_benchmark:
                data[env] = {}
            elif env not in data:
                continue
            for agent, df_agent in df_env.groupby("agent"):
                # All agents or a specific list of agents can be used.
                if baselines != ["all"] and agent not in baselines:
                    continue

                # Rename agents with the same name.
                if agent in data[env]:
                    logger.warning(f"Renaming baseline {agent}")
                    agent = agent + " (baseline)"

                # Extract the data.
                xs, means, stds = [], [], []
                for seed, df_seed in df_agent.groupby("seed"):
                    x = df_seed[x_axis].values
                    if x_min and x[-1] < x_min:
                        logger.warning(
                            f"Skipping {env} {agent} {seed} ({x[-1]} steps)"
                        )
                        continue
                    if y_axis in df_seed:
                        mean = df_seed[y_axis].values
                        if (
                            y_axis[-5:] == "/mean"
                            and y_axis[:-5] + "/std" in df_seed
                        ):
                            std = df_seed[y_axis[:-5] + "/std"].values
                        else:
                            std = None
                    elif y_axis + "/mean" in df_seed:
                        mean = df_seed[y_axis + "/mean"].values
                        if y_axis + "/std" in df_seed:
                            std = df_seed[y_axis + "/std"].values
                        else:
                            std = None
                    else:
                        raise KeyError(f"Key {y_axis} not found.")

                    # Limit the range of x values and smooth.
                    if x_max:
                        max_index = np.argmax(x > x_max) or len(x)
                    else:
                        max_index = len(x)
                    x = x[:max_index]
                    mean = smooth(mean, window)
                    mean = mean[:max_index]
                    if std is not None:
                        std = smooth(std, window)
                        std = std[:max_index]

                    xs.append(x)
                    means.append(mean)
                    stds.append(std)

                if stds[0] is None:
                    stds = None

                print(env, agent)
                data[env][agent] = dict(
                    seeds=(xs, means), stats=stats(xs, means, stds)
                )

    return data


def plot(
    paths,
    x_axis,
    y_axis,
    x_label,
    y_label,
    window,
    interval,
    show_seeds,
    columns,
    x_min,
    x_max,
    y_min,
    y_max,
    baselines,
    baselines_source,
    name,
    save_formats,
    cmap,
    legend_columns,
    legend_marker_size,
    dpi,
    title,
    fig,
):
    """Plots results from experiments and benchmark data."""

    # Extract the data.
    logger.log("Loading data...")
    data = get_data(
        paths,
        baselines,
        baselines_source,
        x_axis,
        y_axis,
        x_min,
        x_max,
        window,
    )

    # List the environments.
    envs = sorted(data.keys(), key=str.casefold)
    num_envs = len(envs)
    if num_envs == 0:
        logger.error("No logs found.")
        return

    # List the agents.
    agents = set()
    for env in data:
        for agent in data[env]:
            agents.add(agent)
    agents = sorted(agents, key=str.casefold)
    num_agents = len(agents)

    if not cmap:
        if num_agents <= 10:
            cmap = "tab10"
        elif num_agents <= 20:
            cmap = "tab20"
        else:
            cmap = "rainbow"
    cmap = plt.get_cmap(cmap)

    if isinstance(cmap, mpl.colors.ListedColormap):
        colors = cmap(range(num_agents))
    else:
        colors = list(cmap(np.linspace(0, 1, num_agents)))

    agent_colors = {a: c for a, c in zip(agents, colors)}
    if columns is None:
        columns = int(np.ceil(np.sqrt(num_envs)))
    else:
        columns = min(columns, num_envs)
    rows = int(np.ceil(num_envs / columns))
    if fig is None:
        plt.ion()
        fig = plt.figure(figsize=(columns * 6, rows * 5))
    else:
        fig.clear()
    grid = gridspec.GridSpec(
        rows + 1,
        1 + columns,
        height_ratios=[1] * rows + [0.1],
        width_ratios=[0] + [1] * columns,
    )
    axes = []
    for i in range(num_envs):
        ax = fig.add_subplot(grid[i // columns, 1 + i % columns])
        axes.append(ax)

    logger.log("Plotting...")
    for env, ax in zip(envs, axes):
        # Plot intervals first.
        if interval in ["std", "bounds"]:
            for agent in sorted(data[env], key=str.casefold):
                color = agent_colors[agent]
                x, mean, min_mean, max_mean, std = data[env][agent]["stats"]
                if interval == "std":
                    if std is None:
                        logger.error("No std found in the data.")
                    else:
                        ax.fill_between(
                            x,
                            mean - std,
                            mean + std,
                            color=color,
                            alpha=0.1,
                            lw=0,
                        )
                elif interval == "bounds":
                    ax.fill_between(
                        x, min_mean, max_mean, color=color, alpha=0.1, lw=0
                    )

        # Plot means after.
        for agent in sorted(data[env], key=str.casefold):
            color = agent_colors[agent]

            # Plot each run if needed.
            xs, means = data[env][agent]["seeds"]
            if show_seeds and len(xs) > 1:
                for x, mean in zip(xs, means):
                    ax.plot(x, mean, c=color, lw=1, alpha=0.5)

            # Plot the overall mean.
            x, mean = data[env][agent]["stats"][:2]
            ax.plot(x, mean, c=color, lw=2, alpha=1)

        # Finalize the figures.
        ax.set_ylim(ymin=y_min, ymax=y_max)
        ax.locator_params(axis="x", nbins=6)
        ax.locator_params(axis="y", tight=True, nbins=6)
        ax.get_yaxis().set_major_formatter(
            mpl.ticker.FuncFormatter(lambda x, p: f"{x:,g}")
        )
        low, high = ax.get_xlim()
        if max(abs(low), abs(high)) >= 1e3:
            ax.ticklabel_format(style="sci", axis="x", scilimits=(0, 0))
        ax.xaxis.grid(linewidth=0.5, alpha=0.5)
        ax.yaxis.grid(linewidth=0.5)
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.tick_params(axis="both", which="both", length=0)
        if x_label is None:
            x_label = "Steps" if x_axis == "train/steps" else x_axis
        if y_label is None:
            y_label = "Score" if y_axis == "test/episode_score" else y_axis
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        if title:
            ax.set_title(title)
        else:
            ax.set_title(env)

    # Add a legend bellow.
    legend_ax = fig.add_subplot(grid[-1:, :])
    legend_ax.set_axis_off()
    handles = []
    for color in colors:
        marker = lines.Line2D(
            range(1),
            range(1),
            marker="o",
            markerfacecolor=color,
            markersize=legend_marker_size,
            linewidth=0,
            markeredgewidth=0,
        )
        handles.append(marker)

    # Find a nice number of legend columns. Labels should not overlap.
    if legend_columns is None:
        legend_columns = range(num_agents, 0, -1)
    else:
        legend_columns = [legend_columns]
    for ncol in legend_columns:
        legend = legend_ax.legend(
            flip(handles, ncol),
            flip(agents, ncol),
            loc="center",
            mode="expand",
            borderaxespad=0,
            borderpad=0,
            handlelength=0.9,
            ncol=ncol,
            numpoints=1,
        )
        legend_frame = legend.get_frame()
        legend_frame.set_linewidth(0)
        fig.tight_layout(pad=0, w_pad=0, h_pad=1.0)
        fig.canvas.draw()
        renderer = legend_ax.get_renderer_cache()
        h_packer = legend.get_children()[0].get_children()[1]
        target_width = h_packer.get_extent(renderer)[0]
        current_width = sum(
            [ch.get_extent(renderer)[0] for ch in h_packer.get_children()]
        )
        if target_width > 1.3 * current_width:
            break

    # Save the plot in different formats if needed.
    if save_formats:
        logger.log("Saving...")
        if name is None:
            if len(envs) > 1:
                name = "results"
            else:
                name = envs[0]
        for save_format in save_formats:
            file_name = name + "." + save_format
            fig.savefig(file_name, facecolor=fig.get_facecolor(), dpi=dpi)
            print(file_name)
        print("to", os.getcwd())

    return fig


def main():
    # Argument parsing.
    parser = argparse.ArgumentParser()
    parser.add_argument("--paths", nargs="+", default=[])
    parser.add_argument("--x_axis", default="train/steps")
    parser.add_argument("--y_axis", default="test/episode_score")
    parser.add_argument("--x_label")
    parser.add_argument("--y_label")
    parser.add_argument("--interval", default="bounds")
    parser.add_argument("--window", type=int, default=1)
    parser.add_argument("--show_seeds", type=bool, default=False)
    parser.add_argument("--columns", type=int)
    parser.add_argument("--x_min", type=int)
    parser.add_argument("--x_max", type=int)
    parser.add_argument("--y_min", type=float)
    parser.add_argument("--y_max", type=float)
    parser.add_argument("--baselines", nargs="+")
    parser.add_argument("--baselines_source", default="tensorflow")
    parser.add_argument("--name")
    parser.add_argument("--save_formats", nargs="*", default=["pdf", "png"])
    parser.add_argument("--seconds", type=int, default=0)
    parser.add_argument("--cmap")
    parser.add_argument("--legend_columns", type=int)
    parser.add_argument("--font_size", type=int, default=12)
    parser.add_argument("--font_family", default="serif")
    parser.add_argument("--legend_font_size", type=int)
    parser.add_argument("--legend_marker_size", type=int, default=10)
    parser.add_argument("--backend", default=None)
    parser.add_argument("--dpi", type=int, default=150)
    parser.add_argument("--title")
    args = parser.parse_args()

    # Backend selection, e.g. agg for non-GUI.
    has_gui = True
    if args.backend:
        mpl.use(args.backend)
        has_gui = args.backend.lower() != "agg"
    del args.backend

    # Fonts.
    plt.rc("font", family=args.font_family, size=args.font_size)
    if args.legend_font_size:
        plt.rc("legend", fontsize=args.legend_font_size)
    del args.font_family, args.font_size, args.legend_font_size

    seconds = args.seconds
    del args.seconds

    # Plot.
    start_time = time.time()
    fig = plot(**vars(args), fig=None)

    try:
        # Wait until the Window is closed if GUI.
        if seconds == 0:
            if has_gui:
                while plt.get_fignums() != []:
                    plt.pause(0.1)

        # Repeatedly plot, waiting a few seconds until interruption.
        else:
            while True:
                if has_gui:
                    while time.time() - start_time < seconds:
                        plt.pause(0.1)
                        assert plt.get_fignums() != []
                else:
                    time.sleep(seconds)
                start_time = time.time()
                plot(**vars(args), fig=fig)

    except Exception:
        pass



import argparse
import os
import time

import yaml

import wandb
from deprl.vendor.tonic import utils


class WandbProcessor:
    def __init__(self, path):
        self._path = path
        self._current_line_number = 0
        self._last_line_number = 0
        self._last_update = os.path.getmtime(path)
        self._setup_wandb()

    def get_line_number(self, data):
        return len(data["train/episode_score/mean"])

    def _setup_wandb(self):
        data = utils.load_csv_to_dict(self._path)
        self._current_line_number = self.get_line_number(data)
        self._log(data)
        self._last_line_number = self.get_line_number(data)

    def _log(self, data):
        for idx in range(self._last_line_number, self._current_line_number):
            logged_data = {k: v[idx] for k, v in data.items()}
            change = 1
            while change:
                for k, v in logged_data.items():
                    if "/mean" in k:
                        logged_data.pop(k)
                        logged_data[k[:-5]] = v
                        change = 1
                        break
                    change = 0
            wandb.log(logged_data, step=idx)

    def update(self):
        self._last_update, updated = utils.check_if_csv_has_updated(
            self._path, self._last_update
        )
        if updated:
            data = utils.load_csv_to_dict(self._path)
            self._current_line_number = self.get_line_number(data)
            self._log(data)
            self._last_line_number = self.get_line_number(data)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--path", type=str)
    parser.add_argument("--username", type=str)
    parser.add_argument("--project", type=str, default="None")
    args = parser.parse_args()
    config = yaml.load(
        open(os.path.join(args.path[:-7], "config.yaml"), "r"),
        Loader=yaml.FullLoader,
    )
    print(args.username)
    wandb.init(project=args.project, entity=args.username, config=config)
    processor = WandbProcessor(args.path)
    while True:
        processor.update()
        time.sleep(100)


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod

import gym
import numpy as np

import deprl  # noqa
from deprl.vendor.tonic import logger


class DummyException(Exception):
    pass


class AbstractWrapper(gym.Wrapper, ABC):
    def merge_args(self, args):
        if args is not None:
            for k, v in args.items():
                setattr(self, k, v)

    def apply_args(self):
        pass

    def render(self, *args, **kwargs):
        pass

    @property
    def force_scale(self):
        if not hasattr(self, "_force_scale"):
            self._force_scale = 0
        return self._force_scale

    @force_scale.setter
    def force_scale(self, force_scale):
        assert force_scale >= 0, f"expected positive value, got {force_scale}"
        self._force_scale = force_scale

    @abstractmethod
    def muscle_lengths(self):
        pass

    @abstractmethod
    def muscle_forces(self):
        pass

    @property
    def tendon_states(self):
        """
        Computes the DEP input. We assume an input
        muscle_length + force_scale * muscle_force
        where force_scale is chosen by the user and the other
        variables are normalized by the encountered max and min
        values.
        """
        lce = self.muscle_lengths()
        f = self.muscle_forces()
        if not hasattr(self, "max_muscle"):
            self.max_muscle = np.zeros_like(lce)
            self.min_muscle = np.ones_like(lce) * 100.0
            self.max_force = -np.ones_like(f) * 100.0
            self.min_force = np.ones_like(f) * 100.0
        if not np.any(np.isnan(lce)):
            self.max_muscle = np.maximum(lce, self.max_muscle)
            self.min_muscle = np.minimum(lce, self.min_muscle)
        if not np.any(np.isnan(f)):
            self.max_force = np.maximum(f, self.max_force)
            self.min_force = np.minimum(f, self.min_force)
        return (
            1.0
            * (
                (
                    (lce - self.min_muscle)
                    / (self.max_muscle - self.min_muscle + 0.1)
                )
                - 0.5
            )
            * 2.0
            + self.force_scale
            * ((f - self.min_force) / (self.max_force - self.min_force + 0.1))
        ).copy()


class ExceptionWrapper(AbstractWrapper):
    """
    Catches MuJoCo related exception thrown mostly by instabilities in the simulation.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def reset(self, **kwargs):
        observation = super().reset(**kwargs)
        if not np.any(np.isnan(observation)):
            self.last_observation = observation.copy()
        else:
            return self.reset(**kwargs)
        return observation

    def step(self, action):
        try:
            observation, reward, done, info = super().step(action)
            if np.any(np.isnan(observation)):
                raise self.error("NaN detected! Resetting.")

        except self.error as e:
            logger.log(f"Simulator exception thrown: {e}")
            observation = self.last_observation
            reward = 0
            done = 1
            info = {}
            self.reset()
        return observation, reward, done, info


class GymWrapper(ExceptionWrapper):
    """Wrapper for OpenAI Gym and MuJoCo, compatible with
    gym=0.13.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        dummy_counter = 0
        try:
            from mujoco_py.builder import MujocoException

            error_mjpy = MujocoException
        except ModuleNotFoundError:
            error_mjpy = DummyException
            dummy_counter += 1
        try:
            from dm_control.rl.control import PhysicsError

            error_mj = PhysicsError
        except ModuleNotFoundError:
            error_mj = DummyException
            dummy_counter += 1

        if dummy_counter >= 2:
            logger.log(
                "Neither mujoco nor mujoco_py has been detected. GymWrapper is not catching exceptions correctly."
            )
        self.error = (error_mjpy, error_mj)

    def render(self, *args, **kwargs):
        kwargs["mode"] = "window"
        self.unwrapped.sim.render(*args, **kwargs)

    def muscle_lengths(self):
        length = self.unwrapped.sim.data.actuator_length
        return length

    def muscle_forces(self):
        return self.unwrapped.sim.data.actuator_force

    def muscle_velocities(self):
        return self.unwrapped.sim.data.actuator_velocity

    def muscle_activity(self):
        return self.unwrapped.sim.data.act

    @property
    def _max_episode_steps(self):
        return self.unwrapped.max_episode_steps


class DMWrapper(ExceptionWrapper):
    """
    Wrapper for general DeepMind ControlSuite environments.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from dm_control.rl.control import PhysicsError

        self.error = PhysicsError

    def muscle_lengths(self):
        length = self.env.environment.physics.data.actuator_length
        return length

    def muscle_forces(self):
        return self.env.environment.physics.data.actuator_force

    def muscle_velocities(self):
        return self.env.environment.physics.data.actuator_velocity

    def muscle_activity(self):
        return self.env.environment.physics.data.act

    @property
    def _max_episode_steps(self):
        return self.unwrapped.max_episode_steps


class OstrichDMWrapper(DMWrapper):
    """
    Wrapper explicitly for the OstrichRL environment.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def muscle_lengths(self):
        return self.env.environment.physics.muscle_lengths().copy()

    def muscle_forces(self):
        return self.env.environment.physics.muscle_forces().copy()

    def muscle_velocities(self):
        return self.env.environment.physics.muscle_velocities().copy()

    def muscle_activity(self):
        return self.env.environment.physics.muscle_activations().copy()


def apply_wrapper(env):
    if "control" in str(type(env)).lower():
        if env.name == "ostrich-run":
            return OstrichDMWrapper(env)
        return DMWrapper(env)
    return GymWrapper(env)


def env_tonic_compat(env, preid=5, parallel=1, sequential=1):
    """
    Applies wrapper for tonic and passes random seed.
    """
    return lambda identifier=0: apply_wrapper(eval(env))



import torch


class ObservationEncoder(torch.nn.Module):
    def initialize(
        self,
        observation_space,
        action_space=None,
        observation_normalizer=None,
    ):
        self.observation_normalizer = observation_normalizer
        observation_size = observation_space.shape[0]
        return observation_size

    def forward(self, observations):
        if self.observation_normalizer:
            observations = self.observation_normalizer(observations)
        return observations


class ObservationActionEncoder(torch.nn.Module):
    def initialize(
        self, observation_space, action_space, observation_normalizer=None
    ):
        self.observation_normalizer = observation_normalizer
        observation_size = observation_space.shape[0]
        action_size = action_space.shape[0]
        return observation_size + action_size

    def forward(self, observations, actions):
        if self.observation_normalizer:
            observations = self.observation_normalizer(observations)
        return torch.cat([observations, actions], dim=-1)



from deprl.vendor.tonic.utils.csv_utils import (
    check_if_csv_has_updated,
    load_csv,
    load_csv_to_dict,
)

from . import logger

__all__ = [logger, load_csv, check_if_csv_has_updated, load_csv_to_dict]



import numpy as np
import torch

FLOAT_EPSILON = 1e-8


def flat_concat(xs):
    return torch.cat([torch.reshape(x, (-1,)) for x in xs], dim=0)


def assign_params_from_flat(new_params, params):
    def flat_size(p):
        return int(np.prod(p.shape))

    splits = torch.split(new_params, [flat_size(p) for p in params])
    new_params = [
        torch.reshape(p_new, p.shape) for p, p_new in zip(params, splits)
    ]
    for p, p_new in zip(params, new_params):
        p.data.copy_(p_new)


class ConjugateGradient:
    def __init__(
        self,
        conjugate_gradient_steps=10,
        damping_coefficient=0.1,
        constraint_threshold=0.01,
        backtrack_steps=10,
        backtrack_coefficient=0.8,
    ):
        self.conjugate_gradient_steps = conjugate_gradient_steps
        self.damping_coefficient = damping_coefficient
        self.constraint_threshold = constraint_threshold
        self.backtrack_steps = backtrack_steps
        self.backtrack_coefficient = backtrack_coefficient

    def optimize(self, loss_function, constraint_function, variables):
        def _hx(x):
            f = constraint_function()
            gradient_1 = torch.autograd.grad(f, variables, create_graph=True)
            gradient_1 = flat_concat(gradient_1)
            x = torch.as_tensor(x)
            y = (gradient_1 * x).sum()
            gradient_2 = torch.autograd.grad(y, variables)
            gradient_2 = flat_concat(gradient_2)

            if self.damping_coefficient > 0:
                gradient_2 += self.damping_coefficient * x

            return gradient_2

        def _cg(b):
            x = np.zeros_like(b)
            r = b.copy()
            p = r.copy()
            r_dot_old = np.dot(r, r)
            if r_dot_old == 0:
                return None

            for _ in range(self.conjugate_gradient_steps):
                z = _hx(p).numpy()
                alpha = r_dot_old / (np.dot(p, z) + FLOAT_EPSILON)
                x += alpha * p
                r -= alpha * z
                r_dot_new = np.dot(r, r)
                p = r + (r_dot_new / r_dot_old) * p
                r_dot_old = r_dot_new
            return x

        def _update(alpha, conjugate_gradient, step, start_variables):
            conjugate_gradient = torch.as_tensor(conjugate_gradient)
            new_variables = start_variables - alpha * conjugate_gradient * step
            assign_params_from_flat(new_variables, variables)
            constraint = constraint_function()
            loss = loss_function()
            return constraint.detach(), loss.detach()

        start_variables = flat_concat(variables)

        for var in variables:
            if var.grad:
                var.grad.data.zero_()

        loss = loss_function()
        grad = torch.autograd.grad(loss, variables)
        grad = flat_concat(grad).numpy()
        start_loss = loss.detach().numpy()

        conjugate_gradient = _cg(grad)
        if conjugate_gradient is None:
            constraint = torch.as_tensor(0.0, dtype=torch.float32)
            loss = torch.as_tensor(0.0, dtype=torch.float32)
            steps = torch.as_tensor(0, dtype=torch.int32)
            return constraint, loss, steps

        alpha = np.sqrt(
            2
            * self.constraint_threshold
            / np.dot(conjugate_gradient, _hx(conjugate_gradient))
            + FLOAT_EPSILON
        )

        if self.backtrack_steps is None or self.backtrack_coefficient is None:
            constraint, loss = _update(
                alpha, conjugate_gradient, 1, start_variables
            )
            return constraint, loss

        for i in range(self.backtrack_steps):
            constraint, loss = _update(
                alpha,
                conjugate_gradient,
                self.backtrack_coefficient**i,
                start_variables,
            )

            if (
                constraint.numpy() <= self.constraint_threshold
                and loss.numpy() <= start_loss
            ):
                break

            if i == self.backtrack_steps - 1:
                constraint, loss = _update(
                    alpha, conjugate_gradient, 0, start_variables
                )
                i = self.backtrack_steps

        return constraint, loss, torch.as_tensor(i + 1, dtype=torch.int32)



from .actor_critics import (
    ActorCritic,
    ActorCriticWithTargets,
    ActorTwinCriticWithTargets,
)
from .actors import (
    Actor,
    DetachedScaleGaussianPolicyHead,
    DeterministicPolicyHead,
    GaussianPolicyHead,
    SquashedMultivariateNormalDiag,
)
from .critics import Critic, DistributionalValueHead, ValueHead
from .encoders import ObservationActionEncoder, ObservationEncoder
from .utils import MLP, default_dense_kwargs

__all__ = [
    default_dense_kwargs,
    MLP,
    ObservationActionEncoder,
    ObservationEncoder,
    SquashedMultivariateNormalDiag,
    DetachedScaleGaussianPolicyHead,
    GaussianPolicyHead,
    DeterministicPolicyHead,
    Actor,
    Critic,
    DistributionalValueHead,
    ValueHead,
    ActorCritic,
    ActorCriticWithTargets,
    ActorTwinCriticWithTargets,
]



import torch

from deprl.vendor.tonic.torch import models, normalizers


def custom_model_mpo(hidden_size=256):
    return models.ActorCriticWithTargets(
        actor=models.Actor(
            encoder=models.ObservationEncoder(),
            torso=models.MLP((hidden_size, hidden_size), torch.nn.ReLU),
            head=models.GaussianPolicyHead(),
        ),
        critic=models.Critic(
            encoder=models.ObservationActionEncoder(),
            torso=models.MLP((hidden_size, hidden_size), torch.nn.ReLU),
            head=models.ValueHead(),
        ),
        observation_normalizer=normalizers.MeanStd(),
    )


def custom_return_mpo(hidden_size=256):
    return models.ActorCriticWithTargets(
        actor=models.Actor(
            encoder=models.ObservationEncoder(),
            torso=models.MLP((hidden_size, hidden_size), torch.nn.ReLU),
            head=models.GaussianPolicyHead(),
        ),
        critic=models.Critic(
            encoder=models.ObservationActionEncoder(),
            torso=models.MLP((hidden_size, hidden_size), torch.nn.ReLU),
            head=models.ValueHead(),
        ),
        observation_normalizer=normalizers.MeanStd(),
        return_normalizer=normalizers.returns.Return(0.99),
    )



import argparse
import json
import os
import sys
from types import SimpleNamespace

import gdown
import yaml

from deprl.vendor.tonic import logger


def prepare_files(orig_params):
    params = get_params(orig_params)
    os.makedirs(params.working_dir, exist_ok=True)
    return params


def get_params(orig_params):
    params = orig_params.copy()
    for key, val in params.items():
        if type(params[key]) == dict:
            params[key] = SimpleNamespace(**val)
    params = SimpleNamespace(**params)
    return params


def prepare_params():
    f = open(sys.argv[-1], "r")
    orig_params = json.load(f)
    params = prepare_files(orig_params)
    return orig_params, params


def load(path, environment, checkpoint="last", noisy=False):
    config, checkpoint_path = load_config_and_paths(path, checkpoint)
    header = config.header
    agent = config.agent
    # Run the header
    exec(header)
    # Build the agent.
    agent = eval(agent)
    # Adapt mpo specific settings
    if "mpo_args" in config:
        agent.set_params(**config.mpo_args)
    # Initialize the agent.
    agent.initialize(
        observation_space=environment.observation_space,
        action_space=environment.action_space,
    )
    # Load the weights of the agent form a checkpoint.
    agent.load(checkpoint_path)
    agent.noisy = noisy
    return agent


def load_baseline(environment):
    identifier = (
        environment.env_name
        if hasattr(environment, "env_name")
        else str(environment)
    )
    print(identifier)
    if "myoLegWalk" in identifier:
        logger.log("Load LegWalk Baseline")
        return load_baseline_myolegwalk(environment)
    if "myoChallengeChaseTagP1" in identifier:
        logger.log("Load ChaseTagP1 Baseline")
        return load_baseline_myochasetagp1(environment, noisy=True)
    if "myoChallengeRelocateP1" in identifier:
        logger.log("Load RelocateP1 Baseline")
        return load_baseline_myorelocatep1(environment)


def load_baseline_myolegwalk(environment, noisy=False):
    modelurl = (
        "https://drive.google.com/uc?id=1UkiUozk-PM8JbQCZNoy2Jr2t1StLcAzx"
    )
    configurl = (
        "https://drive.google.com/uc?id=1knsol05ZL14aqyuaT-TlOvahr31gKQwE"
    )
    foldername = "./baselines_DEPRL/myoLegWalk_20230514/myoLeg"
    if not os.path.exists(foldername):
        os.makedirs(foldername)
        os.makedirs(os.path.join(foldername, "checkpoints"))
    modelpath = os.path.join(foldername, "checkpoints/step_150000000.pt")
    configpath = os.path.join(foldername, "config.yaml")
    if not os.path.exists(modelpath):
        gdown.download(modelurl, modelpath, quiet=False)
        gdown.download(configurl, configpath, quiet=False)
    return load(foldername, environment, noisy=noisy)


def load_baseline_myochasetagp1(environment, noisy=False):
    modelurl = (
        "https://drive.google.com/uc?id=12mEWnwGe7aWzfaHIJT8_qZPINGlPSLfQ"
    )
    configurl = (
        "https://drive.google.com/uc?id=11TRLmNtLMeBQ5H_JZ_tORxl9Idxhq-ec"
    )
    foldername = "./baselines_DEPRL/chasetagp1/"
    if not os.path.exists(foldername):
        os.makedirs(foldername)
        os.makedirs(os.path.join(foldername, "checkpoints"))
    modelpath = os.path.join(foldername, "checkpoints/step_100000000.pt")
    configpath = os.path.join(foldername, "config.yaml")
    if not os.path.exists(modelpath):
        gdown.download(modelurl, modelpath, quiet=False)
        gdown.download(configurl, configpath, quiet=False)
    return load(foldername, environment, noisy=noisy)


def load_baseline_myorelocatep1(environment, noisy=False):
    modelurl = (
        "https://drive.google.com/uc?id=1aBBamewALMxBglkR7nw8gLplKLam3AAO"
    )
    configurl = (
        "https://drive.google.com/uc?id=1UphxBaBLhPplZzhmhZNtkoAcW3M19bmo"
    )
    foldername = "./baselines_DEPRL/relocatep1/"
    if not os.path.exists(foldername):
        os.makedirs(foldername)
        os.makedirs(os.path.join(foldername, "checkpoints"))
    modelpath = os.path.join(foldername, "checkpoints/step_11000000.pt")
    configpath = os.path.join(foldername, "config.yaml")
    if not os.path.exists(modelpath):
        gdown.download(modelurl, modelpath, quiet=False)
        gdown.download(configurl, configpath, quiet=False)
    return load(foldername, environment, noisy=noisy)


def load_config_and_paths(path, checkpoint="last"):
    logger.log(f"Loading experiment from {path}")
    checkpoint_path = os.path.join(path, "checkpoints")
    if not os.path.isdir(checkpoint_path):
        logger.error(f"{checkpoint_path} is not a directory")
        checkpoint_path = None

    # List all the checkpoints.
    checkpoint_ids = []
    for file in os.listdir(checkpoint_path):
        if file[:5] == "step_":
            checkpoint_id = file.split(".")[0]
            checkpoint_ids.append(int(checkpoint_id[5:]))

    if checkpoint_ids:
        # Use the last checkpoint.
        if checkpoint == "last":
            checkpoint_id = max(checkpoint_ids)
            checkpoint_path = os.path.join(
                checkpoint_path, f"step_{checkpoint_id}"
            )

        # Use the specified checkpoint.
        else:
            checkpoint_id = int(checkpoint)
            if checkpoint_id in checkpoint_ids:
                checkpoint_path = os.path.join(
                    checkpoint_path, f"step_{checkpoint_id}"
                )
            else:
                logger.error(
                    f"Checkpoint {checkpoint_id} "
                    f"not found in {checkpoint_path}"
                )
                checkpoint_path = None

    else:
        logger.error(f"No checkpoint found in {checkpoint_path}")
        checkpoint_path = None

    # Load the experiment configuration.
    arguments_path = os.path.join(path, "config.yaml")
    with open(arguments_path, "r") as config_file:
        config = yaml.load(config_file, Loader=yaml.FullLoader)
    config = argparse.Namespace(**config)
    return config, checkpoint_path


def mujoco_render(env, *args, **kwargs):
    if "mujoco_py" in str(type(env.sim)):
        env.render(*args, **kwargs)
    else:
        env.sim.renderer.render_to_window(*args, **kwargs)



"""Script used to train agents."""

import argparse
import os

import tonic
import yaml


def train(
    header,
    agent,
    environment,
    test_environment,
    trainer,
    before_training,
    after_training,
    parallel,
    sequential,
    seed,
    name,
    environment_name,
    checkpoint,
    path,
):
    """Trains an agent on an environment."""

    # Capture the arguments to save them, e.g. to play with the trained agent.
    args = dict(locals())

    checkpoint_path = None

    # Process the checkpoint path same way as in tonic.play
    if path:
        tonic.logger.log(f"Loading experiment from {path}")

        # Use no checkpoint, the agent is freshly created.
        if checkpoint == "none" or agent is not None:
            tonic.logger.log("Not loading any weights")

        else:
            checkpoint_path = os.path.join(path, "checkpoints")
            if not os.path.isdir(checkpoint_path):
                tonic.logger.error(f"{checkpoint_path} is not a directory")
                checkpoint_path = None

            # List all the checkpoints.
            checkpoint_ids = []
            for file in os.listdir(checkpoint_path):
                if file[:5] == "step_":
                    checkpoint_id = file.split(".")[0]
                    checkpoint_ids.append(int(checkpoint_id[5:]))

            if checkpoint_ids:
                # Use the last checkpoint.
                if checkpoint == "last":
                    checkpoint_id = max(checkpoint_ids)
                    checkpoint_path = os.path.join(
                        checkpoint_path, f"step_{checkpoint_id}"
                    )

                # Use the specified checkpoint.
                else:
                    checkpoint_id = int(checkpoint)
                    if checkpoint_id in checkpoint_ids:
                        checkpoint_path = os.path.join(
                            checkpoint_path, f"step_{checkpoint_id}"
                        )
                    else:
                        tonic.logger.error(
                            f"Checkpoint {checkpoint_id} "
                            f"not found in {checkpoint_path}"
                        )
                        checkpoint_path = None

            else:
                tonic.logger.error(f"No checkpoint found in {checkpoint_path}")
                checkpoint_path = None

        # Load the experiment configuration.
        arguments_path = os.path.join(path, "config.yaml")
        with open(arguments_path, "r") as config_file:
            config = yaml.load(config_file, Loader=yaml.FullLoader)
        config = argparse.Namespace(**config)

        header = header or config.header
        agent = agent or config.agent
        environment = environment or config.test_environment
        environment = environment or config.environment
        trainer = trainer or config.trainer

    # Run the header first, e.g. to load an ML framework.
    if header:
        exec(header)

    # Build the training environment.
    _environment = environment
    environment = tonic.environments.distribute(
        lambda: eval(_environment), parallel, sequential
    )
    environment.initialize(seed=seed)

    # Build the testing environment.
    _test_environment = test_environment if test_environment else _environment
    test_environment = tonic.environments.distribute(
        lambda: eval(_test_environment)
    )
    test_environment.initialize(seed=seed + 10000)

    # Build the agent.
    if not agent:
        raise ValueError("No agent specified.")
    agent = eval(agent)
    agent.initialize(
        observation_space=environment.observation_space,
        action_space=environment.action_space,
        seed=seed,
    )

    # Load the weights of the agent form a checkpoint.
    if checkpoint_path:
        agent.load(checkpoint_path)

    # Initialize the logger to save data to the path environment/name/seed.
    if not environment_name:
        if hasattr(test_environment, "name"):
            environment_name = test_environment.name
        else:
            environment_name = test_environment.__class__.__name__
    if not name:
        if hasattr(agent, "name"):
            name = agent.name
        else:
            name = agent.__class__.__name__
        if parallel != 1 or sequential != 1:
            name += f"-{parallel}x{sequential}"
    path = os.path.join(environment_name, name, str(seed))
    tonic.logger.initialize(path, script_path=__file__, config=args)

    # Build the trainer.
    trainer = trainer or "tonic.Trainer()"
    trainer = eval(trainer)
    trainer.initialize(
        agent=agent, environment=environment, test_environment=test_environment
    )

    # Run some code before training.
    if before_training:
        exec(before_training)

    # Train.
    trainer.run()

    # Run some code after training.
    if after_training:
        exec(after_training)


if __name__ == "__main__":
    # Argument parsing.
    parser = argparse.ArgumentParser()
    parser.add_argument("--header")
    parser.add_argument("--agent")
    parser.add_argument("--environment", "--env")
    parser.add_argument("--test_environment", "--test_env")
    parser.add_argument("--trainer")
    parser.add_argument("--before_training")
    parser.add_argument("--after_training")
    parser.add_argument("--parallel", type=int, default=1)
    parser.add_argument("--sequential", type=int, default=1)
    parser.add_argument("--seed", type=int, default=0)
    parser.add_argument("--name")
    parser.add_argument("--environment_name")
    parser.add_argument("--checkpoint", default="last")
    parser.add_argument("--path")

    args = vars(parser.parse_args())
    train(**args)



import torch

FLOAT_EPSILON = 1e-8


class SquashedMultivariateNormalDiag:
    def __init__(self, loc, scale):
        self._distribution = torch.distributions.normal.Normal(loc, scale)

    def rsample_with_log_prob(self, shape=()):
        samples = self._distribution.rsample(shape)
        squashed_samples = torch.tanh(samples)
        log_probs = self._distribution.log_prob(samples)
        log_probs -= torch.log(1 - squashed_samples**2 + 1e-6)
        return squashed_samples, log_probs

    def rsample(self, shape=()):
        samples = self._distribution.rsample(shape)
        return torch.tanh(samples)

    def sample(self, shape=()):
        samples = self._distribution.sample(shape)
        return torch.tanh(samples)

    def log_prob(self, samples):
        """Required unsquashed samples cannot be accurately recovered."""
        raise NotImplementedError(
            "Not implemented to avoid approximation errors. "
            "Use sample_with_log_prob directly."
        )

    @property
    def loc(self):
        return torch.tanh(self._distribution.mean)


class DetachedScaleGaussianPolicyHead(torch.nn.Module):
    def __init__(
        self,
        loc_activation=torch.nn.Tanh,
        loc_fn=None,
        log_scale_init=0.0,
        scale_min=1e-4,
        scale_max=1.0,
        distribution=torch.distributions.normal.Normal,
    ):
        super().__init__()
        self.loc_activation = loc_activation
        self.loc_fn = loc_fn
        self.log_scale_init = log_scale_init
        self.scale_min = scale_min
        self.scale_max = scale_max
        self.distribution = distribution

    def initialize(self, input_size, action_size):
        self.loc_layer = torch.nn.Sequential(
            torch.nn.Linear(input_size, action_size), self.loc_activation()
        )
        if self.loc_fn:
            self.loc_layer.apply(self.loc_fn)
        log_scale = [[self.log_scale_init] * action_size]
        self.log_scale = torch.nn.Parameter(
            torch.as_tensor(log_scale, dtype=torch.float32)
        )

    def forward(self, inputs):
        loc = self.loc_layer(inputs)
        batch_size = inputs.shape[0]
        scale = torch.nn.functional.softplus(self.log_scale) + FLOAT_EPSILON
        scale = torch.clamp(scale, self.scale_min, self.scale_max)
        scale = scale.repeat(batch_size, 1)
        return self.distribution(loc, scale)


class GaussianPolicyHead(torch.nn.Module):
    def __init__(
        self,
        loc_activation=torch.nn.Tanh,
        loc_fn=None,
        scale_activation=torch.nn.Softplus,
        scale_min=1e-4,
        scale_max=1,
        scale_fn=None,
        distribution=torch.distributions.normal.Normal,
    ):
        super().__init__()
        self.loc_activation = loc_activation
        self.loc_fn = loc_fn
        self.scale_activation = scale_activation
        self.scale_min = scale_min
        self.scale_max = scale_max
        self.scale_fn = scale_fn
        self.distribution = distribution

    def initialize(self, input_size, action_size):
        self.loc_layer = torch.nn.Sequential(
            torch.nn.Linear(input_size, action_size), self.loc_activation()
        )
        if self.loc_fn:
            self.loc_layer.apply(self.loc_fn)
        self.scale_layer = torch.nn.Sequential(
            torch.nn.Linear(input_size, action_size), self.scale_activation()
        )
        if self.scale_fn:
            self.scale_layer.apply(self.scale_fn)

    def forward(self, inputs):
        loc = self.loc_layer(inputs)
        scale = self.scale_layer(inputs)
        scale = torch.clamp(scale, self.scale_min, self.scale_max)
        return self.distribution(loc, scale)


class DeterministicPolicyHead(torch.nn.Module):
    def __init__(self, activation=torch.nn.Tanh, fn=None):
        super().__init__()
        self.activation = activation
        self.fn = fn

    def initialize(self, input_size, action_size):
        self.action_layer = torch.nn.Sequential(
            torch.nn.Linear(input_size, action_size), self.activation()
        )
        if self.fn is not None:
            self.action_layer.apply(self.fn)

    def forward(self, inputs):
        return self.action_layer(inputs)


class Actor(torch.nn.Module):
    def __init__(self, encoder, torso, head):
        super().__init__()
        self.encoder = encoder
        self.torso = torso
        self.head = head

    def initialize(
        self, observation_space, action_space, observation_normalizer=None
    ):
        size = self.encoder.initialize(
            observation_space, observation_normalizer
        )
        size = self.torso.initialize(size)
        action_size = action_space.shape[0]
        self.head.initialize(size, action_size)

    def forward(self, *inputs):
        out = self.encoder(*inputs)
        out = self.torso(out)
        return self.head(out)



import numpy as np
import torch


class Return(torch.nn.Module):
    def __init__(self, discount_factor):
        super().__init__()
        assert 0 <= discount_factor < 1
        self.coefficient = 1 / (1 - discount_factor)
        self.min_reward = np.float32(-1)
        self.max_reward = np.float32(1)
        self._low = torch.nn.Parameter(
            torch.as_tensor(
                self.coefficient * self.min_reward, dtype=torch.float32
            ),
            requires_grad=False,
        )
        self._high = torch.nn.Parameter(
            torch.as_tensor(
                self.coefficient * self.max_reward, dtype=torch.float32
            ),
            requires_grad=False,
        )

    def forward(self, val):
        val = torch.sigmoid(val)
        return self._low + val * (self._high - self._low)

    def record(self, values):
        for val in values:
            if val < self.min_reward:
                self.min_reward = np.float32(val)
            elif val > self.max_reward:
                self.max_reward = np.float32(val)

    def update(self):
        self._update(self.min_reward, self.max_reward)

    def _update(self, min_reward, max_reward):
        self._low.data.copy_(
            torch.as_tensor(self.coefficient * min_reward, dtype=torch.float32)
        )
        self._high.data.copy_(
            torch.as_tensor(self.coefficient * max_reward, dtype=torch.float32)
        )



import torch

from deprl.vendor.tonic import replays
from deprl.vendor.tonic.torch import models, normalizers, updaters
from deprl.vendor.tonic.utils import logger

from .agent import Agent


def default_model():
    return models.ActorCritic(
        actor=models.Actor(
            encoder=models.ObservationEncoder(),
            torso=models.MLP((64, 64), torch.nn.Tanh),
            head=models.DetachedScaleGaussianPolicyHead(),
        ),
        critic=models.Critic(
            encoder=models.ObservationEncoder(),
            torso=models.MLP((64, 64), torch.nn.Tanh),
            head=models.ValueHead(),
        ),
        observation_normalizer=normalizers.MeanStd(),
    )


class A2C(Agent):
    """Advantage Actor Critic (aka Vanilla Policy Gradient).
    A3C: https://arxiv.org/pdf/1602.01783.pdf
    """

    def __init__(
        self, model=None, replay=None, actor_updater=None, critic_updater=None
    ):
        self.model = model or default_model()
        self.replay = replay or replays.Segment()
        self.actor_updater = (
            actor_updater or updaters.StochasticPolicyGradient()
        )
        self.critic_updater = critic_updater or updaters.VRegression()

    def initialize(self, observation_space, action_space, seed=None):
        super().initialize(seed=seed)
        self.model.initialize(observation_space, action_space)
        self.replay.initialize(seed)
        self.actor_updater.initialize(self.model)
        self.critic_updater.initialize(self.model)

    def step(self, observations, steps):
        # Sample actions and get their log-probabilities for training.
        actions, log_probs = self._step(observations)
        actions = actions.numpy()
        log_probs = log_probs.numpy()

        # Keep some values for the next update.
        self.last_observations = observations.copy()
        self.last_actions = actions.copy()
        self.last_log_probs = log_probs.copy()

        return actions

    def test_step(self, observations, steps):
        # Sample actions for testing.
        return self._test_step(observations).numpy()

    def update(self, observations, rewards, resets, terminations, steps):
        # Store the last transitions in the replay.
        self.replay.store(
            observations=self.last_observations,
            actions=self.last_actions,
            next_observations=observations,
            rewards=rewards,
            resets=resets,
            terminations=terminations,
            log_probs=self.last_log_probs,
        )

        # Prepare to update the normalizers.
        if self.model.observation_normalizer:
            self.model.observation_normalizer.record(self.last_observations)
        if self.model.return_normalizer:
            self.model.return_normalizer.record(rewards)

        # Update the model if the replay is ready.
        if self.replay.ready():
            self._update()

    def _step(self, observations):
        observations = torch.as_tensor(observations, dtype=torch.float32)
        with torch.no_grad():
            distributions = self.model.actor(observations)
            if hasattr(distributions, "sample_with_log_prob"):
                actions, log_probs = distributions.sample_with_log_prob()
            else:
                actions = distributions.sample()
                log_probs = distributions.log_prob(actions)
            log_probs = log_probs.sum(dim=-1)
        return actions, log_probs

    def _test_step(self, observations):
        observations = torch.as_tensor(observations, dtype=torch.float32)
        with torch.no_grad():
            return self.model.actor(observations).sample()

    def _evaluate(self, observations, next_observations):
        observations = torch.as_tensor(observations, dtype=torch.float32)
        next_observations = torch.as_tensor(
            next_observations, dtype=torch.float32
        )
        with torch.no_grad():
            values = self.model.critic(observations)
            next_values = self.model.critic(next_observations)
        return values, next_values

    def _update(self):
        # Compute the lambda-returns.
        batch = self.replay.get_full("observations", "next_observations")
        values, next_values = self._evaluate(**batch)
        values, next_values = values.numpy(), next_values.numpy()
        self.replay.compute_returns(values, next_values)

        # Update the actor once.
        keys = "observations", "actions", "advantages", "log_probs"
        batch = self.replay.get_full(*keys)
        batch = {k: torch.as_tensor(v) for k, v in batch.items()}
        infos = self.actor_updater(**batch)
        for k, v in infos.items():
            logger.store("actor/" + k, v.numpy())

        # Update the critic multiple times.
        for batch in self.replay.get("observations", "returns"):
            batch = {k: torch.as_tensor(v) for k, v in batch.items()}
            infos = self.critic_updater(**batch)
            for k, v in infos.items():
                logger.store("critic/" + k, v.numpy())

        # Update the normalizers.
        if self.model.observation_normalizer:
            self.model.observation_normalizer.update()
        if self.model.return_normalizer:
            self.model.return_normalizer.update()



import numpy as np

from .dep_controller import DEP
from .vendor.tonic import logger


def dep_factory(mix, instance):
    """
    Class factory returning a class that switches between DEP and the RL agent
    based on one of several heuristics.
    0: UnmixedAgent -> Just RL Agent
    1: InitExploreDEP -> Just DEP exploration to initially fill the buffer.
    2: DetSwitchDEP -> Deterministic switching between DEP and RL agent.
    3: StochSwitchDEP -> Stochastic switching between DEP and RL agent.
    """

    class UnmixedAgent(instance.__class__):
        def step(
            self, observations, steps, tendon_states=None, greedy_episode=None
        ):
            return super().step(observations, steps)

        def test_step(self, observations, steps, tendon_states=None):
            return super().test_step(observations, steps)

        def reset(self):
            pass

        def noisy_test_step(self, observations, steps, tendon_states=None):
            actions = self._step(observations).numpy(force=True)
            self.last_observations = observations.copy()
            self.last_actions = actions.copy()
            return actions

    class BaseDep(instance.__class__):
        """
        Dep basis. Here, DEP exploration is only initially used to pre-fill
        the replay buffer before training.
        """

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.expl = DEP()

        def initialize(self, observation_space, action_space, seed=None):
            super().initialize(observation_space, action_space, seed)
            self.expl.initialize(observation_space, action_space, seed)

        def step(
            self, observations, steps, tendon_states=None, greedy_episode=None
        ):
            if steps > (self.replay.steps_before_batches / 1):
                return super().step(observations, steps)
            actions = self.dep_step(tendon_states, steps)
            self.last_observations = observations.copy()
            self.last_actions = actions.copy()

            return actions

        def test_step(
            self, observations, steps, tendon_states=None, greedy_episode=None
        ):
            return super().test_step(observations, steps)

        def update(self, *args, **kwargs):
            super().update(*args, **kwargs)

        def reset(self):
            pass

        def dep_step(self, tendon_states, steps):
            return self.expl.step(tendon_states, steps)

        def noisy_test_step(self, observations, steps, tendon_states=None):
            actions = self._step(observations).numpy(force=True)
            self.last_observations = observations.copy()
            self.last_actions = actions.copy()
            return actions

    class DetSwitchDep(BaseDep):
        def __init__(self, *args, **kwargs):
            self.switch = 0
            self.since_switch = 1
            return super().__init__(*args, **kwargs)

        def step(
            self, observations, steps, tendon_states=None, greedy_episode=None
        ):
            if steps > (self.replay.steps_before_batches / 1):
                if (
                    self.switch
                    and not self.since_switch % self.expl.intervention_length
                ):
                    self.switch = 0
                    self.since_switch = 1
                if (
                    not self.switch
                    and not self.since_switch % self.expl.rl_length
                ):
                    self.switch = 1
                    self.since_switch = 1
                self.since_switch += 1
                if not self.switch:
                    self.dep_step(tendon_states, steps)
                    return super().step(observations, steps)
            actions = self.dep_step(tendon_states, steps)

            self.last_observations = observations.copy()
            self.last_actions = actions.copy()
            return actions

        def update(self, *args, **kwargs):
            super().update(*args, **kwargs)

    class StochSwitchDep(DetSwitchDep):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # want policy to start first
            self.since_switch = 500

        def step(
            self, observations, steps, tendon_states=None, greedy_episode=False
        ):
            if steps > (self.replay.steps_before_batches / 1):
                if greedy_episode:
                    return super(DetSwitchDep, self).step(
                        observations, steps, tendon_states
                    )
                if self.since_switch > self.expl.intervention_length:
                    # important for Dep to keep learning
                    self.dep_step(tendon_states, steps)
                    if np.random.uniform() < self.expl.intervention_proba:
                        self.since_switch = 0
                    self.since_switch += 1
                    return super(DetSwitchDep, self).step(
                        observations, steps, tendon_states
                    )
            actions = self.dep_step(tendon_states, steps)
            self.last_observations = observations.copy()
            self.last_actions = actions.copy()
            self.since_switch += 1
            return actions

    class PureDep(BaseDep):
        def step(
            self, observations, steps, tendon_states=None, greedy_episode=False
        ):
            if np.any(np.isnan(tendon_states)):
                print("tendon nan!")
            return self.dep_step(tendon_states, steps)

        def update(self, *args, **kwargs):
            pass

        def test_update(self, *args, **kwargs):
            pass

        def test_step(
            self, observations, steps, tendon_states=None, greedy_episode=False
        ):
            # return self.dep_step(tendon_states, steps)[0, :]
            return self.dep_step(tendon_states, steps)

    if mix == 1:
        logger.log("Initial exploration DEP")
        return BaseDep
    elif mix == 2:
        logger.log("Deterministic Switch-DEP.")
        return DetSwitchDep
    elif mix == 3:
        logger.log("Stochastic Switch-DEP. Paper version.")
        return StochSwitchDep
    elif mix == 4:
        logger.log("Pure DEP.")
        return PureDep
    elif mix == 0:
        logger.log("No DEP")
        return UnmixedAgent
    else:
        raise Exception("Invalid agent specified")



import abc


class Agent(abc.ABC):
    """Abstract class used to build agents."""

    def __init__(self, *args, **kwargs):
        self.noisy = False
        super().__init__(*args, **kwargs)

    def initialize(self, observation_space, action_space, seed=None):
        pass

    @abc.abstractmethod
    def step(self, observations, steps):
        """Returns actions during training."""
        pass

    def update(self, observations, rewards, resets, terminations, steps):
        """Informs the agent of the latest transitions during training."""
        pass

    @abc.abstractmethod
    def test_step(self, observations, steps):
        """Returns actions during testing."""
        pass

    def test_update(self, observations, rewards, resets, terminations, steps):
        """Informs the agent of the latest transitions during testing."""
        pass

    def save(self, path):
        """Saves the agent weights during training."""
        pass

    def load(self, path):
        """Reloads the agent weights from a checkpoint."""
        pass

    def __call__(self, observation):
        if not self.noisy:
            return self.test_step(observation, steps=1e6)
        else:
            return self._step(observation).numpy(force=True)



import json
import os
from collections import deque

import gym
import jax.numpy as jnp
import numpy as onp

# jax.config.update('jax_platform_name', 'cpu')


class DEP:
    """
    DEP Implementation from Der et al.(2015).
    Jax is used instead of numpy to speed up computation, GPU strongly
    recommended.
    In the future, proper JAX features such as jit, vmap, etc
    should be used.
    """

    def __init__(self, params_path="default_path"):
        """
        Load default parameters. Should get overwritten by the experiment
        json.
        """
        if params_path == "default_path":
            dirname = os.path.dirname(__file__)
            params_path = os.path.join(
                dirname, "param_files/default_agents.json"
            )

        with open(params_path, "r") as f:
            self.params = json.load(f)["DEP"]

    def initialize(self, observation_space, action_space, seed=None):
        """
        Tonic function that saves action and observation spaces.
        """
        action_space = gym.spaces.Box(
            low=-1, high=1, shape=(action_space.shape)
        )
        self.num_sensors = action_space.shape[0]
        self.num_motors = action_space.shape[0]
        self.n_env = 1

        self.act_scale = self.act_high = action_space.high

        self.action_space = action_space
        self.obs_spec = observation_space.shape
        self.set_params(self.params)

    def step(self, observations, steps=None):
        """
        Main step function. Takes in an observation consisting of
        muscle_lengths + alpha * muscle_forces. Alpha can also be 0.
        """
        if observations.shape != self.obs_spec:
            self.obs_spec = observations.shape
            self._reset(observations.shape)
        if len(observations.shape) == 1:
            observations = observations[jnp.newaxis, :]
        return onp.array(self._get_action(observations)).copy()

    def set_params(self, param_dict):
        for k, v in param_dict.items():
            setattr(self, k, v)
        self._reset()

    def _reset(self, obs_shape=None):
        """
        Reset the controller and setup the buffers and matrices.
        We use a shape of [n_env, num_motors, num_sensors] to have
        a separate DEP matrix per parallel environment, increasing the
        diversity of the exploration data. At the moment,
        it is assumed that num_motors = num_sensors, where the sensors
        directly mirror an actuator state. E.g. joint angle or muscle
        length.
        """
        if obs_shape:
            self.n_env = [obs_shape[0] if len(obs_shape) > 1 else 1][0]
        # Identity model matrix
        self.M = jnp.broadcast_to(
            -jnp.eye(self.num_motors, self.num_sensors),
            (self.n_env, self.num_motors, self.num_sensors),
        )
        # Unnormalized controller matrix
        self.C = jnp.zeros((self.n_env, self.num_motors, self.num_sensors))
        # Normalized controller matrix
        self.C_norm = jnp.zeros(
            (self.n_env, self.num_motors, self.num_sensors)
        )
        # Controller biases
        self.Cb = jnp.zeros((self.n_env, self.num_motors))
        # Filtered observation
        self.obs_smoothed = jnp.zeros((self.n_env, self.num_sensors))
        # Observation and action buffer
        self.buffer = deque(maxlen=self.buffer_size)
        # smoothed_observation
        self.obs_smoothed = jnp.zeros(self.obs_spec)
        # time
        self.t = 0

    def _get_action(self, obs):
        """
        Performs rolling average smoothing on the observation and
        computes a DEP learning step.
        """
        # smoothing
        if self.s4avg > 1 and self.t > 0:
            self.obs_smoothed += (obs - self.obs_smoothed) / self.s4avg
        else:
            self.obs_smoothed = obs

        self.buffer.append([self.obs_smoothed.copy(), None])
        # learning step
        if self.with_learning and len(self.buffer) > (2 + self.time_dist):
            self._learn_controller()
        # new action
        y = self._compute_action()
        self.buffer[-1][1] = y.copy()
        self.t += 1
        return y

    def _q_norm(self, q):
        """
        Normalization function for intermediate action
        obtained by applying the controller matrix to the
        input q = C @ x.
        """
        reg = 10.0 ** (-self.regularization)
        if self.q_norm_selector == "l2":
            q_norm = 1.0 / (jnp.linalg.norm(q, axis=-1) + reg)
        elif self.q_norm_selector == "max":
            q_norm = 1.0 / (max(abs(q), axis=-1) + reg)
        elif self.q_norm_selector == "none":
            q_norm = 1.0
        else:
            raise NotImplementedError(
                "q normalization {self.q_norm_selector} not implemented."
            )

        return q_norm

    def _compute_action(
        self,
    ):
        """
        Compute a DEP action from the current C matrix
        """
        q = jnp.einsum("ijk, ik->ij", self.C_norm, self.obs_smoothed)

        q = jnp.einsum(
            "ij, i->ij",
            q,
            self._q_norm(q),
        )
        y = jnp.maximum(-1, jnp.minimum(1, jnp.tanh(q * self.kappa + self.Cb)))
        y = jnp.einsum("ij, j->ij", y, self.act_scale)
        return y

    def _learn_controller(self):
        """
        Update DEP by one learning step.
        """
        self.C = self._compute_C()
        # linear response in motor space (action -> action)
        R = jnp.einsum("ijk, imk->ijm", self.C, self.M)
        reg = 10.0 ** (-self.regularization)
        # controller normalization c.f. Der et al (2015).
        if self.normalization == "independent":
            factor = self.kappa / (jnp.linalg.norm(R, axis=-1) + reg)
            self.C_norm = jnp.einsum("ijk,ik->ijk", self.C, factor)
        elif self.normalization == "none":
            self.C_norm = self.C
        elif self.normalization == "global":
            norm = jnp.linalg.norm(R)
            self.C_norm = self.C * self.kappa / (norm + reg)
        else:
            raise NotImplementedError(
                f"Controller matrix normalization {self.normalization} not implemented."
            )

        if self.bias_rate >= 0:
            yy = self.buffer[-2][1]
            self.Cb -= (
                jnp.clip(yy * self.bias_rate, -0.05, 0.05) + self.Cb * 0.001
            )
        else:
            self.Cb *= 0

    def _compute_C(self):
        """
        Recompute the controller matrix C from the
        buffer of recent transitions. This is similar
        to the rolling average shown in the publication, but without
        recency weighting.
        """
        C = jnp.zeros_like(self.C)
        for s in range(2, min(self.t - self.time_dist, self.tau)):
            x = self.buffer[-s][0][:, : self.num_sensors]
            xx = self.buffer[-s - 1][0][:, : self.num_sensors]
            xx_t = (
                x
                if self.time_dist == 0
                else self.buffer[-s - self.time_dist][0][:, : self.num_sensors]
            )
            xxx_t = self.buffer[-s - 1 - self.time_dist][0][
                :, : self.num_sensors
            ]

            chi = x - xx
            v = xx_t - xxx_t
            mu = jnp.einsum("ijk, ik->ij", self.M, chi)

            C += jnp.einsum("ij, ik->ijk", mu, v)
        return C



import numpy as np
import tensorflow as tf


class Return(tf.keras.Model):
    def __init__(self, discount_factor):
        super().__init__(name="reward_normalizer")
        assert 0 <= discount_factor < 1
        self.coefficient = 1 / (1 - discount_factor)
        self.min_reward = np.float32(-1)
        self.max_reward = np.float32(1)
        self._low = tf.Variable(
            self.coefficient * self.min_reward,
            dtype=tf.float32,
            trainable=False,
            name="low",
        )
        self._high = tf.Variable(
            self.coefficient * self.max_reward,
            dtype=np.float32,
            trainable=False,
            name="high",
        )

    def call(self, val):
        val = tf.sigmoid(val)
        return self._low + val * (self._high - self._low)

    def record(self, values):
        for val in values:
            if val < self.min_reward:
                self.min_reward = np.float32(val)
            elif val > self.max_reward:
                self.max_reward = np.float32(val)

    # Careful: do not use in @tf.function
    def update(self):
        self._update(self.min_reward, self.max_reward)

    @tf.function
    def _update(self, min_reward, max_reward):
        self._low.assign(self.coefficient * min_reward)
        self._high.assign(self.coefficient * max_reward)



from .a2c import A2C  # noqa
from .agent import Agent
from .d4pg import D4PG  # noqa
from .ddpg import DDPG
from .mpo import MPO
from .ppo import PPO
from .sac import SAC
from .td3 import TD3
from .td4 import TD4
from .trpo import TRPO

__all__ = [Agent, A2C, DDPG, D4PG, MPO, PPO, SAC, TD3, TD4, TRPO]



import tensorflow as tf
from tonic import logger
from tonic.tensorflow import agents, updaters


class PPO(agents.A2C):
    """Proximal Policy Optimization.
    PPO: https://arxiv.org/pdf/1707.06347.pdf
    """

    def __init__(
        self, model=None, replay=None, actor_updater=None, critic_updater=None
    ):
        actor_updater = actor_updater or updaters.ClippedRatio()
        super().__init__(
            model=model,
            replay=replay,
            actor_updater=actor_updater,
            critic_updater=critic_updater,
        )

    def _update(self):
        # Compute the lambda-returns.
        batch = self.replay.get_full("observations", "next_observations")
        values, next_values = self._evaluate(**batch)
        values, next_values = values.numpy(), next_values.numpy()
        self.replay.compute_returns(values, next_values)

        train_actor = True
        actor_iterations = 0
        critic_iterations = 0
        keys = "observations", "actions", "advantages", "log_probs", "returns"

        # Update both the actor and the critic multiple times.
        for batch in self.replay.get(*keys):
            if train_actor:
                infos = self._update_actor_critic(**batch)
                actor_iterations += 1
            else:
                batch = {k: batch[k] for k in ("observations", "returns")}
                infos = dict(critic=self.critic_updater(**batch))
            critic_iterations += 1

            # Stop earlier the training of the actor.
            if train_actor:
                train_actor = not infos["actor"]["stop"].numpy()

            for key in infos:
                for k, v in infos[key].items():
                    logger.store(key + "/" + k, v.numpy())

        logger.store("actor/iterations", actor_iterations)
        logger.store("critic/iterations", critic_iterations)

        # Update the normalizers.
        if self.model.observation_normalizer:
            self.model.observation_normalizer.update()
        if self.model.return_normalizer:
            self.model.return_normalizer.update()

    @tf.function
    def _update_actor_critic(
        self, observations, actions, advantages, log_probs, returns
    ):
        actor_infos = self.actor_updater(
            observations, actions, advantages, log_probs
        )
        critic_infos = self.critic_updater(observations, returns)
        return dict(actor=actor_infos, critic=critic_infos)



import numpy as np


def lambda_returns(
    values,
    next_values,
    rewards,
    resets,
    terminations,
    discount_factor,
    trace_decay,
):
    """Function used to calculate lambda-returns on parallel buffers."""

    returns = np.zeros_like(values)
    last_returns = next_values[-1]
    for t in reversed(range(len(rewards))):
        bootstrap = (1 - trace_decay) * next_values[
            t
        ] + trace_decay * last_returns
        bootstrap *= 1 - resets[t]
        bootstrap += resets[t] * next_values[t]
        bootstrap *= 1 - terminations[t]
        returns[t] = last_returns = rewards[t] + discount_factor * bootstrap
    return returns


def flatten_batch(values):
    shape = values.shape
    new_shape = (np.prod(shape[:2], dtype=int),) + shape[2:]
    return values.reshape(new_shape)



import torch

from deprl.vendor.tonic.torch import models, updaters  # noqa

FLOAT_EPSILON = 1e-8


class StochasticPolicyGradient:
    def __init__(self, optimizer=None, entropy_coeff=0, gradient_clip=0):
        self.optimizer = optimizer or (
            lambda params: torch.optim.Adam(params, lr=3e-4)
        )
        self.entropy_coeff = entropy_coeff
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = models.trainable_variables(self.model.actor)
        self.optimizer = self.optimizer(self.variables)

    def __call__(self, observations, actions, advantages, log_probs):
        if (advantages == 0.0).all():
            loss = torch.as_tensor(0.0, dtype=torch.float32)
            kl = torch.as_tensor(0.0, dtype=torch.float32)
            with torch.no_grad():
                distributions = self.model.actor(observations)
                entropy = distributions.entropy().mean()
                std = distributions.stddev.mean()

        else:
            self.optimizer.zero_grad()
            distributions = self.model.actor(observations)
            new_log_probs = distributions.log_prob(actions).sum(dim=-1)
            loss = -(advantages * new_log_probs).mean()
            entropy = distributions.entropy().mean()
            if self.entropy_coeff != 0:
                loss -= self.entropy_coeff * entropy

            loss.backward()
            if self.gradient_clip > 0:
                torch.nn.utils.clip_grad_norm_(
                    self.variables, self.gradient_clip
                )
            self.optimizer.step()

            loss = loss.detach()
            kl = (log_probs - new_log_probs).mean().detach()
            entropy = entropy.detach()
            std = distributions.stddev.mean().detach()

        return dict(loss=loss, kl=kl, entropy=entropy, std=std)


class ClippedRatio:
    def __init__(
        self,
        optimizer=None,
        ratio_clip=0.2,
        kl_threshold=0.015,
        entropy_coeff=0,
        gradient_clip=0,
    ):
        self.optimizer = optimizer or (
            lambda params: torch.optim.Adam(params, lr=3e-4)
        )
        self.ratio_clip = ratio_clip
        self.kl_threshold = kl_threshold
        self.entropy_coeff = entropy_coeff
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = models.trainable_variables(self.model.actor)
        self.optimizer = self.optimizer(self.variables)

    def __call__(self, observations, actions, advantages, log_probs):
        if (advantages == 0.0).all():
            loss = torch.as_tensor(0.0, dtype=torch.float32)
            kl = torch.as_tensor(0.0, dtype=torch.float32)
            clip_fraction = torch.as_tensor(0.0, dtype=torch.float32)
            with torch.no_grad():
                distributions = self.model.actor(observations)
                entropy = distributions.entropy().mean()
                std = distributions.stddev.mean()

        else:
            self.optimizer.zero_grad()
            distributions = self.model.actor(observations)
            new_log_probs = distributions.log_prob(actions).sum(dim=-1)
            ratios_1 = torch.exp(new_log_probs - log_probs)
            surrogates_1 = advantages * ratios_1
            ratio_low = 1 - self.ratio_clip
            ratio_high = 1 + self.ratio_clip
            ratios_2 = torch.clamp(ratios_1, ratio_low, ratio_high)
            surrogates_2 = advantages * ratios_2
            loss = -(torch.min(surrogates_1, surrogates_2)).mean()
            entropy = distributions.entropy().mean()
            if self.entropy_coeff != 0:
                loss -= self.entropy_coeff * entropy

            loss.backward()
            if self.gradient_clip > 0:
                torch.nn.utils.clip_grad_norm_(
                    self.variables, self.gradient_clip
                )
            self.optimizer.step()

            loss = loss.detach()
            with torch.no_grad():
                kl = (log_probs - new_log_probs).mean()
            entropy = entropy.detach()
            clipped = ratios_1.gt(ratio_high) | ratios_1.lt(ratio_low)
            clip_fraction = torch.as_tensor(
                clipped, dtype=torch.float32
            ).mean()
            std = distributions.stddev.mean().detach()

        return dict(
            loss=loss,
            kl=kl,
            entropy=entropy,
            clip_fraction=clip_fraction,
            std=std,
            stop=kl > self.kl_threshold,
        )


class TrustRegionPolicyGradient:
    def __init__(self, optimizer=None, entropy_coeff=0):
        self.optimizer = optimizer or updaters.ConjugateGradient()
        self.entropy_coeff = entropy_coeff

    def initialize(self, model):
        self.model = model
        self.variables = models.trainable_variables(self.model.actor)

    def __call__(
        self, observations, actions, log_probs, locs, scales, advantages
    ):
        if (advantages == 0.0).all():
            kl = torch.as_tensor(0.0, dtype=torch.float32)
            loss = torch.as_tensor(0.0, dtype=torch.float32)
            steps = torch.as_tensor(0, dtype=torch.int32)

        else:
            kl, loss, steps = self.optimizer.optimize(
                loss_function=lambda: self._loss(
                    observations, actions, log_probs, advantages
                ),
                constraint_function=lambda: self._kl(
                    observations, locs, scales
                ),
                variables=self.variables,
            )

        return dict(loss=loss, kl=kl, backtrack_steps=steps)

    def _loss(self, observations, actions, old_log_probs, advantages):
        distributions = self.model.actor(observations)
        log_probs = distributions.log_prob(actions).sum(dim=-1)
        ratios = torch.exp(log_probs - old_log_probs)
        loss = -(ratios * advantages).mean()
        if self.entropy_coeff != 0:
            entropy = distributions.entropy().mean()
            loss -= self.entropy_coeff * entropy
        return loss

    def _kl(self, observations, locs, scales):
        distributions = self.model.actor(observations)
        old_distributions = type(distributions)(locs, scales)
        return torch.distributions.kl.kl_divergence(
            distributions, old_distributions
        ).mean()


class DeterministicPolicyGradient:
    def __init__(self, optimizer=None, gradient_clip=0):
        self.optimizer = optimizer or (
            lambda params: torch.optim.Adam(params, lr=1e-3)
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = models.trainable_variables(self.model.actor)
        self.optimizer = self.optimizer(self.variables)

    def __call__(self, observations):
        critic_variables = models.trainable_variables(self.model.critic)

        for var in critic_variables:
            var.requires_grad = False

        self.optimizer.zero_grad()
        actions = self.model.actor(observations)
        values = self.model.critic(observations, actions)
        loss = -values.mean()

        loss.backward()
        if self.gradient_clip > 0:
            torch.nn.utils.clip_grad_norm_(self.variables, self.gradient_clip)
        self.optimizer.step()

        for var in critic_variables:
            var.requires_grad = True

        return dict(loss=loss.detach())


class DistributionalDeterministicPolicyGradient:
    def __init__(self, optimizer=None, gradient_clip=0):
        self.optimizer = optimizer or (
            lambda params: torch.optim.Adam(params, lr=1e-3)
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = models.trainable_variables(self.model.actor)
        self.optimizer = self.optimizer(self.variables)

    def __call__(self, observations):
        critic_variables = models.trainable_variables(self.model.critic)

        for var in critic_variables:
            var.requires_grad = False

        self.optimizer.zero_grad()
        actions = self.model.actor(observations)
        value_distributions = self.model.critic(observations, actions)
        values = value_distributions.mean()
        loss = -values.mean()

        loss.backward()
        if self.gradient_clip > 0:
            torch.nn.utils.clip_grad_norm_(self.variables, self.gradient_clip)
        self.optimizer.step()

        for var in critic_variables:
            var.requires_grad = True

        return dict(loss=loss.detach())


class TwinCriticSoftDeterministicPolicyGradient:
    def __init__(self, optimizer=None, entropy_coeff=0.2, gradient_clip=0):
        self.optimizer = optimizer or (
            lambda params: torch.optim.Adam(params, lr=3e-4)
        )
        self.entropy_coeff = entropy_coeff
        self.gradient_clip = gradient_clip

    def initialize(self, model):
        self.model = model
        self.variables = models.trainable_variables(self.model.actor)
        self.optimizer = self.optimizer(self.variables)

    def __call__(self, observations):
        critic_1_variables = models.trainable_variables(self.model.critic_1)
        critic_2_variables = models.trainable_variables(self.model.critic_2)
        critic_variables = critic_1_variables + critic_2_variables

        for var in critic_variables:
            var.requires_grad = False

        self.optimizer.zero_grad()
        distributions = self.model.actor(observations)
        if hasattr(distributions, "rsample_with_log_prob"):
            actions, log_probs = distributions.rsample_with_log_prob()
        else:
            actions = distributions.rsample()
            log_probs = distributions.log_prob(actions)
        log_probs = log_probs.sum(dim=-1)
        values_1 = self.model.critic_1(observations, actions)
        values_2 = self.model.critic_2(observations, actions)
        values = torch.min(values_1, values_2)
        loss = (self.entropy_coeff * log_probs - values).mean()

        loss.backward()
        if self.gradient_clip > 0:
            torch.nn.utils.clip_grad_norm_(self.variables, self.gradient_clip)
        self.optimizer.step()

        for var in critic_variables:
            var.requires_grad = True

        return dict(loss=loss.detach())


class MaximumAPosterioriPolicyOptimization:
    def __init__(
        self,
        num_samples=20,
        epsilon=1e-1,
        epsilon_penalty=1e-3,
        epsilon_mean=1e-3,
        epsilon_std=1e-6,
        initial_log_temperature=1.0,
        initial_log_alpha_mean=1.0,
        initial_log_alpha_std=10.0,
        min_log_dual=-18.0,
        per_dim_constraining=True,
        action_penalization=True,
        actor_optimizer=None,
        dual_optimizer=None,
        gradient_clip=0,
    ):
        self.num_samples = num_samples
        self.epsilon = epsilon
        self.epsilon_mean = epsilon_mean
        self.epsilon_std = epsilon_std
        self.initial_log_temperature = initial_log_temperature
        self.initial_log_alpha_mean = initial_log_alpha_mean
        self.initial_log_alpha_std = initial_log_alpha_std
        self.min_log_dual = torch.as_tensor(min_log_dual, dtype=torch.float32)
        self.action_penalization = action_penalization
        self.epsilon_penalty = epsilon_penalty
        self.per_dim_constraining = per_dim_constraining
        self.actor_optimizer = actor_optimizer or (
            lambda params: torch.optim.Adam(params, lr=3e-4)
        )
        self.dual_optimizer = actor_optimizer or (
            lambda params: torch.optim.Adam(params, lr=1e-2)
        )
        self.gradient_clip = gradient_clip

    def initialize(self, model, action_space):
        self.model = model
        self.actor_variables = models.trainable_variables(self.model.actor)
        self.actor_optimizer = self.actor_optimizer(self.actor_variables)

        # Dual variables.
        self.dual_variables = []
        self.log_temperature = torch.nn.Parameter(
            torch.as_tensor(
                [self.initial_log_temperature], dtype=torch.float32
            )
        )
        self.dual_variables.append(self.log_temperature)
        shape = [action_space.shape[0]] if self.per_dim_constraining else [1]
        self.log_alpha_mean = torch.nn.Parameter(
            torch.full(shape, self.initial_log_alpha_mean, dtype=torch.float32)
        )
        self.dual_variables.append(self.log_alpha_mean)
        self.log_alpha_std = torch.nn.Parameter(
            torch.full(shape, self.initial_log_alpha_std, dtype=torch.float32)
        )
        self.dual_variables.append(self.log_alpha_std)
        if self.action_penalization:
            self.log_penalty_temperature = torch.nn.Parameter(
                torch.as_tensor(
                    [self.initial_log_temperature], dtype=torch.float32
                )
            )
            self.dual_variables.append(self.log_penalty_temperature)
        self.dual_optimizer = self.dual_optimizer(self.dual_variables)

    def __call__(self, observations):
        def parametric_kl_and_dual_losses(kl, alpha, epsilon):
            kl_mean = kl.mean(dim=0)
            kl_loss = (alpha.detach() * kl_mean).sum()
            alpha_loss = (alpha * (epsilon - kl_mean.detach())).sum()
            return kl_loss, alpha_loss

        def weights_and_temperature_loss(q_values, epsilon, temperature):
            tempered_q_values = q_values.detach() / temperature
            weights = torch.nn.functional.softmax(tempered_q_values, dim=0)
            weights = weights.detach()

            # Temperature loss (dual of the E-step).
            q_log_sum_exp = torch.logsumexp(tempered_q_values, dim=0)
            num_actions = torch.as_tensor(
                q_values.shape[0], dtype=torch.float32
            )
            log_num_actions = torch.log(num_actions)
            loss = epsilon + (q_log_sum_exp).mean() - log_num_actions
            loss = temperature * loss

            return weights, loss

        # Use independent normals to satisfy KL constraints per-dimension.
        def independent_normals(distribution_1, distribution_2=None):
            distribution_2 = distribution_2 or distribution_1
            return torch.distributions.independent.Independent(
                torch.distributions.normal.Normal(
                    distribution_1.mean, distribution_2.stddev
                ),
                -1,
            )

        with torch.no_grad():
            self.log_temperature.data.copy_(
                torch.maximum(self.min_log_dual, self.log_temperature)
            )
            self.log_alpha_mean.data.copy_(
                torch.maximum(self.min_log_dual, self.log_alpha_mean)
            )
            self.log_alpha_std.data.copy_(
                torch.maximum(self.min_log_dual, self.log_alpha_std)
            )
            if self.action_penalization:
                self.log_penalty_temperature.data.copy_(
                    torch.maximum(
                        self.min_log_dual, self.log_penalty_temperature
                    )
                )

            target_distributions = self.model.target_actor(observations)
            actions = target_distributions.sample((self.num_samples,))

            tiled_observations = updaters.tile(observations, self.num_samples)
            flat_observations = updaters.merge_first_two_dims(
                tiled_observations
            )
            flat_actions = updaters.merge_first_two_dims(actions)
            values = self.model.target_critic(flat_observations, flat_actions)
            values = values.view(self.num_samples, -1)

            assert isinstance(
                target_distributions, torch.distributions.normal.Normal
            )
            target_distributions = independent_normals(target_distributions)

        self.actor_optimizer.zero_grad()
        self.dual_optimizer.zero_grad()

        distributions = self.model.actor(observations)
        distributions = independent_normals(distributions)

        temperature = (
            torch.nn.functional.softplus(self.log_temperature) + FLOAT_EPSILON
        )
        alpha_mean = (
            torch.nn.functional.softplus(self.log_alpha_mean) + FLOAT_EPSILON
        )
        alpha_std = (
            torch.nn.functional.softplus(self.log_alpha_std) + FLOAT_EPSILON
        )
        weights, temperature_loss = weights_and_temperature_loss(
            values, self.epsilon, temperature
        )

        # Action penalization is quadratic beyond [-1, 1].
        if self.action_penalization:
            penalty_temperature = (
                torch.nn.functional.softplus(self.log_penalty_temperature)
                + FLOAT_EPSILON
            )
            diff_bounds = actions - torch.clamp(actions, -1, 1)
            action_bound_costs = -torch.norm(diff_bounds, dim=-1)
            (
                penalty_weights,
                penalty_temperature_loss,
            ) = weights_and_temperature_loss(
                action_bound_costs, self.epsilon_penalty, penalty_temperature
            )
            weights += penalty_weights
            temperature_loss += penalty_temperature_loss

        # Decompose the policy into fixed-mean and fixed-std distributions.
        fixed_std_distribution = independent_normals(
            distributions.base_dist, target_distributions.base_dist
        )
        fixed_mean_distribution = independent_normals(
            target_distributions.base_dist, distributions.base_dist
        )

        # Compute the decomposed policy losses.
        policy_mean_losses = (
            fixed_std_distribution.base_dist.log_prob(actions).sum(dim=-1)
            * weights
        ).sum(dim=0)
        policy_mean_loss = -(policy_mean_losses).mean()
        policy_std_losses = (
            fixed_mean_distribution.base_dist.log_prob(actions).sum(dim=-1)
            * weights
        ).sum(dim=0)
        policy_std_loss = -policy_std_losses.mean()

        # Compute the decomposed KL between the target and online policies.
        if self.per_dim_constraining:
            kl_mean = torch.distributions.kl.kl_divergence(
                target_distributions.base_dist,
                fixed_std_distribution.base_dist,
            )
            kl_std = torch.distributions.kl.kl_divergence(
                target_distributions.base_dist,
                fixed_mean_distribution.base_dist,
            )
        else:
            kl_mean = torch.distributions.kl.kl_divergence(
                target_distributions, fixed_std_distribution
            )
            kl_std = torch.distributions.kl.kl_divergence(
                target_distributions, fixed_mean_distribution
            )

        # Compute the alpha-weighted KL-penalty and dual losses.
        kl_mean_loss, alpha_mean_loss = parametric_kl_and_dual_losses(
            kl_mean, alpha_mean, self.epsilon_mean
        )
        kl_std_loss, alpha_std_loss = parametric_kl_and_dual_losses(
            kl_std, alpha_std, self.epsilon_std
        )

        # Combine losses.
        policy_loss = policy_mean_loss + policy_std_loss
        kl_loss = kl_mean_loss + kl_std_loss
        dual_loss = alpha_mean_loss + alpha_std_loss + temperature_loss
        loss = policy_loss + kl_loss + dual_loss

        loss.backward()
        if self.gradient_clip > 0:
            torch.nn.utils.clip_grad_norm_(
                self.actor_variables, self.gradient_clip
            )
            torch.nn.utils.clip_grad_norm_(
                self.dual_variables, self.gradient_clip
            )
        self.actor_optimizer.step()
        self.dual_optimizer.step()

        dual_variables = dict(
            temperature=temperature.detach(),
            alpha_mean=alpha_mean.detach(),
            alpha_std=alpha_std.detach(),
        )
        if self.action_penalization:
            dual_variables[
                "penalty_temperature"
            ] = penalty_temperature.detach()

        return dict(
            policy_mean_loss=policy_mean_loss.detach(),
            policy_std_loss=policy_std_loss.detach(),
            kl_mean_loss=kl_mean_loss.detach(),
            kl_std_loss=kl_std_loss.detach(),
            alpha_mean_loss=alpha_mean_loss.detach(),
            alpha_std_loss=alpha_std_loss.detach(),
            temperature_loss=temperature_loss.detach(),
            **dual_variables,
        )



"""Script used to play with trained agents."""

import argparse
import os

import numpy as np
import yaml

from deprl import env_wrappers, mujoco_render

from .vendor.tonic import logger


def play_gym(agent, environment):
    """Launches an agent in a Gym-based environment."""
    environment = env_wrappers.apply_wrapper(environment)
    observations = environment.reset()
    tendon_states = environment.tendon_states

    score = 0
    length = 0
    min_reward = float("inf")
    max_reward = -float("inf")
    global_min_reward = float("inf")
    global_max_reward = -float("inf")
    steps = 0
    episodes = 0

    while True:
        actions = agent.test_step(
            observations, tendon_states=tendon_states, steps=1e6
        )
        if len(actions.shape) > 1:
            actions = actions[0, :]
        observations, reward, done, info = environment.step(actions)
        tendon_states = environment.tendon_states
        mujoco_render(environment)

        steps += 1
        score += reward
        min_reward = min(min_reward, reward)
        max_reward = max(max_reward, reward)
        global_min_reward = min(global_min_reward, reward)
        global_max_reward = max(global_max_reward, reward)
        length += 1

        if done or length >= environment.max_episode_steps:
            episodes += 1

            print()
            print(f"Episodes: {episodes:,}")
            print(f"Score: {score:,.3f}")
            print(f"Length: {length:,}")
            print(f"Terminal: {done:}")
            print(f"Min reward: {min_reward:,.3f}")
            print(f"Max reward: {max_reward:,.3f}")
            print(f"Global min reward: {min_reward:,.3f}")
            print(f"Global max reward: {max_reward:,.3f}")
            environment.reset()

            score = 0
            length = 0
            min_reward = float("inf")
            max_reward = -float("inf")


def play_control_suite(agent, environment):
    """Launches an agent in a DeepMind Control Suite-based environment."""

    from dm_control import viewer

    class Wrapper:
        """Wrapper used to plug a Tonic environment in a dm_control viewer."""

        def __init__(self, environment):
            self.environment = environment
            self.unwrapped = environment.unwrapped
            self.action_spec = self.unwrapped.environment.action_spec
            self.physics = self.unwrapped.environment.physics
            self.infos = None
            self.steps = 0
            self.episodes = 0
            self.min_reward = float("inf")
            self.max_reward = -float("inf")
            self.global_min_reward = float("inf")
            self.global_max_reward = -float("inf")
            self.max_vel = 0

        def reset(self):
            """Mimics a dm_control reset for the viewer."""

            self.observations = self.environment.reset()[None]

            self.score = 0
            self.length = 0
            self.min_reward = float("inf")
            self.max_reward = -float("inf")

            return self.unwrapped.last_time_step

        def step(self, actions):
            """Mimics a dm_control step for the viewer."""

            assert not np.isnan(actions.sum())
            ob, rew, term, _ = self.environment.step(actions[0])

            self.score += rew
            self.length += 1
            self.min_reward = min(self.min_reward, rew)
            self.max_reward = max(self.max_reward, rew)
            self.global_min_reward = min(self.global_min_reward, rew)
            self.global_max_reward = max(self.global_max_reward, rew)
            timeout = self.length == self.environment.max_episode_steps
            done = term or timeout

            if done:
                self.episodes += 1
                print()
                print(f"Episodes: {self.episodes:,}")
                print(f"Score: {self.score:,.3f}")
                print(f"Length: {self.length:,}")
                print(f"Terminal: {term:}")
                print(f"Min reward: {self.min_reward:,.3f}")
                print(f"Max reward: {self.max_reward:,.3f}")
                print(f"Global min reward: {self.min_reward:,.3f}")
                print(f"Global max reward: {self.max_reward:,.3f}")

            self.observations = ob[None]
            self.infos = dict(
                observations=ob[None],
                rewards=np.array([rew]),
                resets=np.array([done]),
                terminations=np.array([term]),
            )

            return self.unwrapped.last_time_step

        @property
        def tendon_states(self):
            return self.environment.tendon_states

    # Wrap the environment for the viewer.
    environment = env_wrappers.apply_wrapper(environment)
    environment = Wrapper(environment)

    def policy(timestep):
        """Mimics a dm_control policy for the viewer."""
        if environment.infos is not None:
            agent.test_update(**environment.infos, steps=environment.steps)
            environment.steps += 1
        tendon_states = environment.tendon_states
        return agent.test_step(
            environment.observations,
            tendon_states=tendon_states,
            steps=environment.steps,
        )

    # Launch the viewer with the wrapped environment and policy.
    viewer.launch(environment, policy)


def play(path, checkpoint, seed, header, agent, environment):
    """Reloads an agent and an environment from a previous experiment."""

    checkpoint_path = None

    if path:
        logger.log(f"Loading experiment from {path}")

        # Use no checkpoint, the agent is freshly created.
        if checkpoint == "none" or agent is not None:
            logger.log("Not loading any weights")

        else:
            checkpoint_path = os.path.join(path, "checkpoints")
            if not os.path.isdir(checkpoint_path):
                logger.error(f"{checkpoint_path} is not a directory")
                checkpoint_path = None

            # List all the checkpoints.
            checkpoint_ids = []
            for file in os.listdir(checkpoint_path):
                if file[:5] == "step_":
                    checkpoint_id = file.split(".")[0]
                    checkpoint_ids.append(int(checkpoint_id[5:]))

            if checkpoint_ids:
                # Use the last checkpoint.
                if checkpoint == "last":
                    checkpoint_id = max(checkpoint_ids)
                    checkpoint_path = os.path.join(
                        checkpoint_path, f"step_{checkpoint_id}"
                    )

                # Use the specified checkpoint.
                else:
                    checkpoint_id = int(checkpoint)
                    if checkpoint_id in checkpoint_ids:
                        checkpoint_path = os.path.join(
                            checkpoint_path, f"step_{checkpoint_id}"
                        )
                    else:
                        logger.error(
                            f"Checkpoint {checkpoint_id} "
                            f"not found in {checkpoint_path}"
                        )
                        checkpoint_path = None

            else:
                logger.error(f"No checkpoint found in {checkpoint_path}")
                checkpoint_path = None

        # Load the experiment configuration.
        arguments_path = os.path.join(path, "config.yaml")
        with open(arguments_path, "r") as config_file:
            config = yaml.load(config_file, Loader=yaml.FullLoader)
        config = argparse.Namespace(**config)

        header = header or config.header
        agent = agent or config.agent
        environment = environment or config.test_environment
        environment = environment or config.environment

    # Run the header first, e.g. to load an ML framework.
    if header:
        exec(header)

    # Build the agent.
    if not agent:
        raise ValueError("No agent specified.")
    agent = eval(agent)

    # Build the environment.
    environment = eval(environment)
    environment.seed(seed)

    # Adapt mpo specific settings
    if "config" in locals():
        if "mpo_args" in config:
            agent.set_params(**config.mpo_args)
    # Initialize the agent.
    agent.initialize(
        observation_space=environment.observation_space,
        action_space=environment.action_space,
        seed=seed,
    )

    # Load the weights of the agent form a checkpoint.
    if checkpoint_path:
        agent.load(checkpoint_path)
    from pudb import set_trace; set_trace()


if __name__ == "__main__":
    # Argument parsing.
    parser = argparse.ArgumentParser()
    parser.add_argument("--path")
    parser.add_argument("--checkpoint", default="last")
    parser.add_argument("--seed", type=int, default=0)
    parser.add_argument("--header")
    parser.add_argument("--agent")
    parser.add_argument("--environment", "--env")
    args = vars(parser.parse_args())
    play(**args)



"""Script used to play with trained agents."""
import argparse
import os

import numpy as np
import deprl  # noqa
import yaml
from deprl import env_tonic_compat
# from catatonic.utils import relabel_and_get_trifinger_reward
from matplotlib import pyplot as plt

from deprl.utils import ScoreKeeper
from deprl.utils.utils_scone_gait_sto import new_get_avg_dtwd

# from .utils import env_tonic_compat, replace_reward_trifinger

def trafo_myoleg(gait):
    return gait
    for i in range(gait.shape[-1]):
        gait[:, i] = np.roll(gait[:, i], gait.shape[0]//2)

    return gait

def convert_data_myoleg(gaits, gaits_mean):
    # trafo = lambda arr: np.roll(arr, arr.shape[0] // 2, axis=0)
    # trafo = lambda arr: np.roll(arr, 10, axis=0)
    # trafo2 = lambda x: -x
    # trafo = lambda x:
    trafo2 = lambda x: x

    for gait in gaits:
        for i in range(len(gait)):
            gait[i] = trafo_myoleg(gait[i])
            gait[i][:, 1] = trafo2((gait[i][:, 1]))
    for i in range(len(gaits_mean)):
        gaits_mean[i] = trafo_myoleg(gaits_mean[i])
        gaits_mean[i][:, 1] = trafo2((gaits_mean[i][:, 1]))
    return gaits, gaits_mean

def get_qpos(environment, joint_name):
    adr = environment.sim.model.jnt_qposadr[
        environment.sim.model.joint_name2id(joint_name)
    ]
    return environment.sim.data.qpos[adr]

def get_grf(env):
    sensors = ['l_toes', 'r_toes', 'l_foot', 'r_foot']
    weight = np.sum(env.sim.model.body_mass[:16]) * 10
    sensor_val = np.array([env.sim.data.get_sensor(sensor) / weight for sensor in sensors])
    return sensor_val.copy()


def get_grf(env):
    sensors = ['l_toes', 'l_foot', 'r_toes', 'r_foot']
    weight = np.sum(env.sim.model.body_mass[:16]) * 9.81
    sensor_val = np.array([env.sim.data.get_sensor(sensor) / weight for sensor in sensors])
    sensor_val[0] = np.sum(sensor_val[:2])
    sensor_val[1] = np.sum(sensor_val[2:])
    return sensor_val[:2].copy()


def get_joints(env):
    joint_names = ['hip_flexion_l', 'knee_angle_l', 'ankle_angle_l', 'hip_flexion_r', 'knee_angle_r', 'ankle_angle_r']
    joint_val = np.array([get_qpos(env, joint_name) for joint_name in joint_names])
    return joint_val.copy()

def play_gym(agent, environment, env_args=None):
    """Launches an agent in a Gym-based environment."""
    # environment = tonic.environments.distribute(
    # lambda identifier=0: environment, env_args=env_args
    # )
    observations = environment.reset()
    tendon_states = environment.tendon_states
    score_keeper = ScoreKeeper()
    score = 0
    length = 0
    min_reward = float("inf")
    max_reward = -float("inf")
    global_min_reward = float("inf")
    global_max_reward = -float("inf")
    steps = 0
    episodes = 0
    maxes = np.zeros_like(environment.action_space.shape)
    grf = []
    joints = []
    acts = []

    while True:
        actions = agent.test_step(observations, steps, tendon_states)
        # actions = agent.noisy_step(observations, steps, tendon_states)
        # actions[:] = -1
        observations, reward, done, infos = environment.step(actions)
        tendon_states = environment.tendon_states
        if hasattr(environment.sim, "data"):
            maxes = np.maximum(maxes, environment.sim.data.qfrc_actuator)
        steps += 1
        score += reward
        if hasattr(environment, "rwd_dict"):
            grf.append(get_grf(environment))
            joints.append(get_joints(environment))
            acts.append(environment.sim.data.act.copy())
        min_reward = min(min_reward, reward)
        max_reward = max(max_reward, reward)
        global_min_reward = min(global_min_reward, reward)
        global_max_reward = max(global_max_reward, reward)
        length += 1
        if done or length >= 2000:
            term = done
            episodes += 1

            print()
            print(f"Episodes: {episodes:,}")
            print(f"Score: {score:,.3f}")
            print(f"Length: {length:,}")
            print(f"Terminal: {term:}")
            print(f"Min reward: {min_reward:,.3f}")
            print(f"Max reward: {max_reward:,.3f}")
            print(f"Global min reward: {min_reward:,.3f}")
            print(f"Global max reward: {max_reward:,.3f}")
            observations = environment.reset()
            tendon_states = environment.tendon_states
            joints, acts, grf = np.array(joints), np.array(acts), np.array(grf)
            os.makedirs('./recorded_data', exist_ok=True)
            if length > 500:
                np.save(f'./recorded_data/myoleg_efficient_joints_ep_{episodes}.npy', joints)
                np.save(f'./recorded_data/myoleg_efficient_acts_ep_{episodes}.npy', acts)
                np.save(f'./recorded_data/myoleg_efficient_grf_ep_{episodes}.npy', grf)

            time = np.arange(joints.shape[0]) * 0.01
            joints = np.hstack([time[:, np.newaxis], joints])
            dtwd_sum, dtwd_hip, dtwd_knee, dtwd_ankle = new_get_avg_dtwd([joints], [grf],   plot=False, convert_fn=convert_data_myoleg)
            print(f'{dtwd_sum=} {dtwd_hip=} {dtwd_knee=} {dtwd_ankle=}')
            grf = []
            joints = []
            acts = []

            score = 0
            length = 0
            success_counter = 0
            overall_success = 0
            effort = []
            min_reward = float("inf")
            max_reward = -float("inf")
            # break
        if episodes > 20:
            if "sconegym" in str(type(environment)):
                environment.model.write_results("sconepy_example")
            break


def play(path, checkpoint, seed, header, agent, environment):
    """Reloads an agent and an environment from a previous experiment."""

    checkpoint_path = None

    if path:
        deprl.logger.log(f"Loading experiment from {path}")

        # Use no checkpoint, the agent is freshly created.
        if checkpoint == "none" or agent is not None:
            deprl.logger.log("Not loading any weights")

        else:
            checkpoint_path = os.path.join(path, "checkpoints")
            if not os.path.isdir(checkpoint_path):
                deprl.logger.error(f"{checkpoint_path} is not a directory")
                checkpoint_path = None

            # List all the checkpoints.
            checkpoint_ids = []
            for file in os.listdir(checkpoint_path):
                if file[:5] == "step_":
                    checkpoint_id = file.split(".")[0]
                    checkpoint_ids.append(int(checkpoint_id[5:]))

            if checkpoint_ids:
                # Use the last checkpoint.
                if checkpoint == "last":
                    checkpoint_id = max(checkpoint_ids)
                    checkpoint_path = os.path.join(
                        checkpoint_path, f"step_{checkpoint_id}"
                    )

                # Use the specified checkpoint.
                else:
                    checkpoint_id = int(checkpoint)
                    if checkpoint_id in checkpoint_ids:
                        checkpoint_path = os.path.join(
                            checkpoint_path, f"step_{checkpoint_id}"
                        )
                    else:
                        deprl.logger.error(
                            f"Checkpoint {checkpoint_id} not found in {checkpoint_path}"
                        )
                        checkpoint_path = None

            else:
                deprl.logger.error(f"No checkpoint found in {checkpoint_path}")
                checkpoint_path = None

        # Load the experiment configuration.
        arguments_path = os.path.join(path, "config.yaml")
        with open(arguments_path, "r") as config_file:
            config = yaml.load(config_file, Loader=yaml.FullLoader)
        config = argparse.Namespace(**config)
        print(config)
        header = header or config.header
        agent = agent or config.agent
        environment = environment or config.test_environment
        environment = environment or config.environment

    # Run the header first, e.g. to load an ML framework.
    if header:
        exec(header)

    # Build the agent.
    if not agent:
        raise ValueError("No agent specified.")
    agent = eval(agent)

    # Build the environment.
    environment = env_tonic_compat(environment)()
    environment.seed(seed)

    # Adapt mpo specific settings
    if "config" in locals():
        if "mpo_args" in config:
            agent.set_params(**config.mpo_args)

    # Initialize the agent.
    agent.initialize(
        observation_space=environment.observation_space,
        action_space=environment.action_space,
        seed=seed,
    )

    # Load the weights of the agent form a checkpoint.
    if checkpoint_path:
        agent.load(checkpoint_path)

    # Play with the agent in the environment.
    if isinstance(environment, deprl.environments.wrappers.ActionRescaler):
        environment_type = environment.env.__class__.__name__
    else:
        environment_type = environment.__class__.__name__
    if "config" in locals() and hasattr(config, "env_args"):
        play_gym(agent, environment, config.env_args)
    else:
        play_gym(agent, environment)


if __name__ == "__main__":
    # Argument parsing.
    parser = argparse.ArgumentParser()
    parser.add_argument("--path")
    parser.add_argument("--checkpoint", default="last")
    parser.add_argument("--seed", type=int, default=0)
    parser.add_argument("--header")
    parser.add_argument("--agent")
    parser.add_argument("--environment", "--env")
    args = vars(parser.parse_args())
    play(**args)



import torch

from deprl import custom_torso
from deprl.vendor.tonic.torch import agents, updaters


class TunedMPO(agents.MPO):
    """Maximum a Posteriori Policy Optimisation.
    MPO: https://arxiv.org/pdf/1806.06920.pdf
    MO-MPO: https://arxiv.org/pdf/2005.07513.pdf
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def set_params(
        self,
        lr_critic=1e-3,
        grad_clip_critic=0,
        lr_actor=3e-4,
        lr_dual=1e-2,
        grad_clip_actor=0,
        hidden_size=None,
        batch_size=None,
        retnorm=None,
        return_steps=None,
    ):
        def optim_critic(params):
            return torch.optim.Adam(params, lr_critic)

        self.critic_updater = TunedExpectedSARSA(
            optimizer=optim_critic, gradient_clip=grad_clip_critic
        )

        def optim_actor(params):
            return torch.optim.Adam(params, lr=lr_actor)

        def optim_dual(params):
            return torch.optim.Adam(params, lr=lr_dual)

        self.actor_updater = TunedMaximumAPosteriori(
            actor_optimizer=optim_actor,
            dual_optimizer=optim_dual,
            gradient_clip=grad_clip_actor,
        )
        if hidden_size is None:
            hidden_size = 256
        if retnorm is not None:
            self.model = custom_torso.custom_return_mpo(
                hidden_size=hidden_size
            )
        else:
            self.model = custom_torso.custom_model_mpo(hidden_size=hidden_size)
        if batch_size is not None:
            self.replay.batch_size = batch_size
        if return_steps is not None:
            self.replay.return_steps = return_steps

    def initialize(self, *args, **kwargs):
        super().initialize(*args, **kwargs)


class TunedExpectedSARSA(updaters.critics.ExpectedSARSA):
    def __init__(
        self, num_samples=20, loss=None, optimizer=None, gradient_clip=0
    ):
        self.num_samples = num_samples
        self.loss = loss or torch.nn.MSELoss()
        self.optimizer = optimizer or (
            lambda params: torch.optim.Adam(params, lr=3e-4, weight_decay=1e-5)
        )
        self.gradient_clip = gradient_clip


class TunedMaximumAPosteriori(
    updaters.actors.MaximumAPosterioriPolicyOptimization
):
    def __init__(
        self,
        num_samples=20,
        epsilon=1e-1,
        epsilon_penalty=1e-3,
        epsilon_mean=1e-3,
        epsilon_std=1e-6,
        initial_log_temperature=1.0,
        initial_log_alpha_mean=1.0,
        initial_log_alpha_std=10.0,
        min_log_dual=-18.0,
        per_dim_constraining=True,
        action_penalization=True,
        actor_optimizer=None,
        dual_optimizer=None,
        gradient_clip=0,
    ):
        self.num_samples = num_samples
        self.epsilon = epsilon
        self.epsilon_mean = epsilon_mean
        self.epsilon_std = epsilon_std
        self.initial_log_temperature = initial_log_temperature
        self.initial_log_alpha_mean = initial_log_alpha_mean
        self.initial_log_alpha_std = initial_log_alpha_std
        self.min_log_dual = torch.as_tensor(min_log_dual, dtype=torch.float32)
        self.action_penalization = action_penalization
        self.epsilon_penalty = epsilon_penalty
        self.per_dim_constraining = per_dim_constraining
        self.actor_optimizer = actor_optimizer or (
            lambda params: torch.optim.Adam(params, lr=3e-4, weight_decay=1e-5)
        )
        self.dual_optimizer = dual_optimizer or (
            lambda params: torch.optim.Adam(params, lr=1e-2, weight_decay=1e-5)
        )
        self.gradient_clip = gradient_clip



import torch


class ValueHead(torch.nn.Module):
    def __init__(self, fn=None):
        super().__init__()
        self.fn = fn

    def initialize(self, input_size, return_normalizer=None):
        self.return_normalizer = return_normalizer
        self.v_layer = torch.nn.Linear(input_size, 1)
        if self.fn:
            self.v_layer.apply(self.fn)

    def forward(self, inputs):
        out = self.v_layer(inputs)
        out = torch.squeeze(out, -1)
        if self.return_normalizer:
            out = self.return_normalizer(out)
        return out


class CategoricalWithSupport:
    def __init__(self, values, logits):
        self.values = values
        self.logits = logits
        self.probabilities = torch.nn.functional.softmax(logits, dim=-1)

    def mean(self):
        return (self.probabilities * self.values).sum(dim=-1)

    def project(self, returns):
        vmin, vmax = self.values[0], self.values[-1]
        d_pos = torch.cat([self.values, vmin[None]], 0)[1:]
        d_pos = (d_pos - self.values)[None, :, None]
        d_neg = torch.cat([vmax[None], self.values], 0)[:-1]
        d_neg = (self.values - d_neg)[None, :, None]

        clipped_returns = torch.clamp(returns, vmin, vmax)
        delta_values = clipped_returns[:, None] - self.values[None, :, None]
        delta_sign = (delta_values >= 0).float()
        delta_hat = (delta_sign * delta_values / d_pos) - (
            (1 - delta_sign) * delta_values / d_neg
        )
        delta_clipped = torch.clamp(1 - delta_hat, 0, 1)

        return (delta_clipped * self.probabilities[:, None]).sum(dim=2)


class DistributionalValueHead(torch.nn.Module):
    def __init__(self, vmin, vmax, num_atoms, fn=None):
        super().__init__()
        self.num_atoms = num_atoms
        self.fn = fn
        self.values = torch.linspace(vmin, vmax, num_atoms).float()

    def initialize(self, input_size, return_normalizer=None):
        if return_normalizer:
            raise ValueError(
                "Return normalizers cannot be used with distributional value"
                "heads."
            )
        self.distributional_layer = torch.nn.Linear(input_size, self.num_atoms)
        if self.fn:
            self.distributional_layer.apply(self.fn)

    def forward(self, inputs):
        logits = self.distributional_layer(inputs)
        return CategoricalWithSupport(values=self.values, logits=logits)


class Critic(torch.nn.Module):
    def __init__(self, encoder, torso, head):
        super().__init__()
        self.encoder = encoder
        self.torso = torso
        self.head = head

    def initialize(
        self,
        observation_space,
        action_space,
        observation_normalizer=None,
        return_normalizer=None,
    ):
        size = self.encoder.initialize(
            observation_space=observation_space,
            action_space=action_space,
            observation_normalizer=observation_normalizer,
        )
        size = self.torso.initialize(size)
        self.head.initialize(size, return_normalizer)

    def forward(self, *inputs):
        out = self.encoder(*inputs)
        out = self.torso(out)
        return self.head(out)



from .builders import Bullet, ControlSuite, Gym
from .distributed import Parallel, Sequential, distribute
from .wrappers import ActionRescaler, TimeFeature

__all__ = [
    Bullet,
    ControlSuite,
    Gym,
    distribute,
    Parallel,
    Sequential,
    ActionRescaler,
    TimeFeature,
]



import numpy as np


class Buffer:
    """Replay storing a large number of transitions for off-policy learning
    and using n-step returns."""

    def __init__(
        self,
        size=int(1e6),
        return_steps=1,
        batch_iterations=50,
        batch_size=100,
        discount_factor=0.99,
        steps_before_batches=int(1e4),
        steps_between_batches=50,
    ):
        self.full_max_size = size
        self.return_steps = return_steps
        self.batch_iterations = batch_iterations
        self.batch_size = batch_size
        self.discount_factor = discount_factor
        self.steps_before_batches = steps_before_batches
        self.steps_between_batches = steps_between_batches

    def initialize(self, seed=None):
        self.np_random = np.random.RandomState(seed)
        self.buffers = None
        self.index = 0
        self.size = 0
        self.last_steps = 0

    def ready(self, steps):
        if steps < self.steps_before_batches:
            return False
        return (steps - self.last_steps) >= self.steps_between_batches

    def store(self, **kwargs):
        if "terminations" in kwargs:
            continuations = np.float32(1 - kwargs["terminations"])
            kwargs["discounts"] = continuations * self.discount_factor

        # Create the named buffers.
        if self.buffers is None:
            self.num_workers = len(list(kwargs.values())[0])
            self.max_size = self.full_max_size // self.num_workers
            self.buffers = {}
            for key, val in kwargs.items():
                shape = (self.max_size,) + np.array(val).shape
                self.buffers[key] = np.full(shape, np.nan, np.float32)

        # Store the new values.
        for key, val in kwargs.items():
            self.buffers[key][self.index] = val

        # Accumulate values for n-step returns.
        if self.return_steps > 1:
            self.accumulate_n_steps(kwargs)

        self.index = (self.index + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def accumulate_n_steps(self, kwargs):
        rewards = kwargs["rewards"]
        next_observations = kwargs["next_observations"]
        discounts = kwargs["discounts"]
        masks = np.ones(self.num_workers, np.float32)

        for i in range(min(self.size, self.return_steps - 1)):
            index = (self.index - i - 1) % self.max_size
            masks *= 1 - self.buffers["resets"][index]
            new_rewards = (
                self.buffers["rewards"][index]
                + self.buffers["discounts"][index] * rewards
            )
            self.buffers["rewards"][index] = (1 - masks) * self.buffers[
                "rewards"
            ][index] + masks * new_rewards
            new_discounts = self.buffers["discounts"][index] * discounts
            self.buffers["discounts"][index] = (1 - masks) * self.buffers[
                "discounts"
            ][index] + masks * new_discounts
            self.buffers["next_observations"][index] = (1 - masks)[
                :, None
            ] * self.buffers["next_observations"][index] + masks[
                :, None
            ] * next_observations

    def get(self, *keys, steps):
        """Get batches from named buffers."""

        for _ in range(self.batch_iterations):
            total_size = self.size * self.num_workers
            indices = self.np_random.randint(total_size, size=self.batch_size)
            rows = indices // self.num_workers
            columns = indices % self.num_workers
            yield {k: self.buffers[k][rows, columns] for k in keys}

        self.last_steps = steps



from deprl.utils.analysis_utils import ScoreKeeper
from deprl.utils.utils import (
    load,
    load_baseline,
    mujoco_render,
    prepare_params,
)

__all__ = [prepare_params, load, load_baseline, ScoreKeeper, mujoco_render]



"""Non-differentiable noisy exploration methods."""

import numpy as np


class NoActionNoise:
    def __init__(self, start_steps=20000):
        self.start_steps = start_steps

    def initialize(self, policy, action_space, seed=None):
        self.policy = policy
        self.action_size = action_space.shape[0]
        self.np_random = np.random.RandomState(seed)

    def __call__(self, observations, steps):
        if steps > self.start_steps:
            actions = self.policy(observations)
            actions = np.clip(actions, -1, 1)
        else:
            shape = (len(observations), self.action_size)
            actions = self.np_random.uniform(-1, 1, shape)
        return actions

    def update(self, resets):
        pass


class NormalActionNoise:
    def __init__(self, scale=0.1, start_steps=20000):
        self.scale = scale
        self.start_steps = start_steps

    def initialize(self, policy, action_space, seed=None):
        self.policy = policy
        self.action_size = action_space.shape[0]
        self.np_random = np.random.RandomState(seed)

    def __call__(self, observations, steps):
        if steps > self.start_steps:
            actions = self.policy(observations)
            noises = self.scale * self.np_random.normal(size=actions.shape)
            actions = (actions + noises).astype(np.float32)
            actions = np.clip(actions, -1, 1)
        else:
            shape = (len(observations), self.action_size)
            actions = self.np_random.uniform(-1, 1, shape)
        return actions

    def update(self, resets):
        pass


class OrnsteinUhlenbeckActionNoise:
    def __init__(
        self, scale=0.1, clip=2, theta=0.15, dt=1e-2, start_steps=20000
    ):
        self.scale = scale
        self.clip = clip
        self.theta = theta
        self.dt = dt
        self.start_steps = start_steps

    def initialize(self, policy, action_space, seed=None):
        self.policy = policy
        self.action_size = action_space.shape[0]
        self.np_random = np.random.RandomState(seed)
        self.noises = None

    def __call__(self, observations, steps):
        if steps > self.start_steps:
            actions = self.policy(observations)

            if self.noises is None:
                self.noises = np.zeros_like(actions)
            noises = self.np_random.normal(size=actions.shape)
            noises = np.clip(noises, -self.clip, self.clip)
            self.noises -= self.theta * self.noises * self.dt
            self.noises += self.scale * np.sqrt(self.dt) * noises
            actions = (actions + self.noises).astype(np.float32)
            actions = np.clip(actions, -1, 1)
        else:
            shape = (len(observations), self.action_size)
            actions = self.np_random.uniform(-1, 1, shape)
        return actions

    def update(self, resets):
        if self.noises is not None:
            self.noises *= (1.0 - resets)[:, None]



import copy

import tensorflow as tf


class ActorCritic(tf.keras.Model):
    def __init__(
        self,
        actor,
        critic,
        observation_normalizer=None,
        return_normalizer=None,
    ):
        super().__init__()
        self.actor = actor
        self.critic = critic
        self.observation_normalizer = observation_normalizer
        self.return_normalizer = return_normalizer

    def initialize(self, observation_space, action_space):
        if self.observation_normalizer:
            self.observation_normalizer.initialize(observation_space.shape)
        self.actor.initialize(
            observation_space, action_space, self.observation_normalizer
        )
        self.critic.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )
        dummy_observations = tf.zeros((1,) + observation_space.shape)
        self.actor(dummy_observations)
        self.critic(dummy_observations)


class ActorCriticWithTargets(tf.keras.Model):
    def __init__(
        self,
        actor,
        critic,
        observation_normalizer=None,
        return_normalizer=None,
        target_coeff=0.005,
    ):
        super().__init__()
        self.actor = actor
        self.critic = critic
        self.target_actor = copy.deepcopy(actor)
        self.target_critic = copy.deepcopy(critic)
        self.observation_normalizer = observation_normalizer
        self.return_normalizer = return_normalizer
        self.target_coeff = target_coeff

    def initialize(self, observation_space, action_space):
        if self.observation_normalizer:
            self.observation_normalizer.initialize(observation_space.shape)
        self.actor.initialize(
            observation_space, action_space, self.observation_normalizer
        )
        self.critic.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )
        self.target_actor.initialize(
            observation_space, action_space, self.observation_normalizer
        )
        self.target_critic.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )
        dummy_observations = tf.zeros((1,) + observation_space.shape)
        dummy_actions = tf.zeros((1,) + action_space.shape)
        self.actor(dummy_observations)
        self.critic(dummy_observations, dummy_actions)
        self.target_actor(dummy_observations)
        self.target_critic(dummy_observations, dummy_actions)
        self.online_variables = (
            self.actor.trainable_variables + self.critic.trainable_variables
        )
        self.target_variables = (
            self.target_actor.trainable_variables
            + self.target_critic.trainable_variables
        )
        self.assign_targets()

    def assign_targets(self):
        for o, t in zip(self.online_variables, self.target_variables):
            t.assign(o)

    def update_targets(self):
        for o, t in zip(self.online_variables, self.target_variables):
            t.assign((1 - self.target_coeff) * t + self.target_coeff * o)


class ActorTwinCriticWithTargets(tf.keras.Model):
    def __init__(
        self,
        actor,
        critic,
        observation_normalizer=None,
        return_normalizer=None,
        target_coeff=0.005,
    ):
        super().__init__()
        self.actor = actor
        self.critic_1 = critic
        self.critic_2 = copy.deepcopy(critic)
        self.target_actor = copy.deepcopy(actor)
        self.target_critic_1 = copy.deepcopy(critic)
        self.target_critic_2 = copy.deepcopy(critic)
        self.observation_normalizer = observation_normalizer
        self.return_normalizer = return_normalizer
        self.target_coeff = target_coeff

    def initialize(self, observation_space, action_space):
        if self.observation_normalizer:
            self.observation_normalizer.initialize(observation_space.shape)
        self.actor.initialize(
            observation_space, action_space, self.observation_normalizer
        )
        self.critic_1.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )
        self.critic_2.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )
        self.target_actor.initialize(
            observation_space, action_space, self.observation_normalizer
        )
        self.target_critic_1.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )
        self.target_critic_2.initialize(
            observation_space,
            action_space,
            self.observation_normalizer,
            self.return_normalizer,
        )
        dummy_observations = tf.zeros((1,) + observation_space.shape)
        dummy_actions = tf.zeros((1,) + action_space.shape)
        self.actor(dummy_observations)
        self.critic_1(dummy_observations, dummy_actions)
        self.critic_2(dummy_observations, dummy_actions)
        self.target_actor(dummy_observations)
        self.target_critic_1(dummy_observations, dummy_actions)
        self.target_critic_2(dummy_observations, dummy_actions)
        self.online_variables = (
            self.actor.trainable_variables
            + self.critic_1.trainable_variables
            + self.critic_2.trainable_variables
        )
        self.target_variables = (
            self.target_actor.trainable_variables
            + self.target_critic_1.trainable_variables
            + self.target_critic_2.trainable_variables
        )
        self.assign_targets()

    def assign_targets(self):
        for o, t in zip(self.online_variables, self.target_variables):
            t.assign(o)

    def update_targets(self):
        for o, t in zip(self.online_variables, self.target_variables):
            t.assign((1 - self.target_coeff) * t + self.target_coeff * o)
